#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""Unit tests for taps"""
import unittest
import yaml
from . import acl
from .unittestyaml import UnitTestYaml


class TestACLMethods(unittest.TestCase):
    def setUp(self):
        with UnitTestYaml("test_acl.yaml") as f:
            self.cfg = yaml.load(f, Loader=yaml.FullLoader)

    def test_get_acls(self):
        acllist = acl.get_acls(self.cfg)
        self.assertIsInstance(acllist, list)
        self.assertEqual(2, len(acllist))

    def test_get_by_name(self):
        aclname, _acl = acl.get_by_name(self.cfg, "deny-all")
        self.assertIsNotNone(_acl)
        self.assertEqual("deny-all", aclname)

        aclname, _acl = acl.get_by_name(self.cfg, "acl-noexist")
        self.assertIsNone(aclname)
        self.assertIsNone(_acl)

    def test_get_port_low_high(self):
        lo, hi = acl.get_port_low_high(80)
        self.assertEqual(80, lo)
        self.assertEqual(80, hi)

        lo, hi = acl.get_port_low_high("80")
        self.assertEqual(80, lo)
        self.assertEqual(80, hi)

        lo, hi = acl.get_port_low_high("www")
        self.assertEqual(80, lo)
        self.assertEqual(80, hi)

        lo, hi = acl.get_port_low_high("any")
        self.assertEqual(0, lo)
        self.assertEqual(65535, hi)

        lo, hi = acl.get_port_low_high("-1024")
        self.assertEqual(0, lo)
        self.assertEqual(1024, hi)

        lo, hi = acl.get_port_low_high("1024-")
        self.assertEqual(1024, lo)
        self.assertEqual(65535, hi)

        lo, hi = acl.get_port_low_high("1000-2000")
        self.assertEqual(1000, lo)
        self.assertEqual(2000, hi)

        lo, hi = acl.get_port_low_high("0-65535")
        self.assertEqual(0, lo)
        self.assertEqual(65535, hi)

        lo, hi = acl.get_port_low_high("bla")
        self.assertIsNone(lo)
        self.assertIsNone(hi)

        lo, hi = acl.get_port_low_high("foo-bar")
        self.assertIsNone(lo)
        self.assertIsNone(hi)

    def test_get_protocol(self):
        proto = acl.get_protocol(1)
        self.assertEqual(1, proto)

        proto = acl.get_protocol("icmp")
        self.assertEqual(1, proto)

        proto = acl.get_protocol("unknown")
        self.assertIsNone(proto)

    def test_get_icmp_low_high(self):
        lo, hi = acl.get_icmp_low_high(3)
        self.assertEqual(3, lo)
        self.assertEqual(3, hi)

        lo, hi = acl.get_icmp_low_high("3")
        self.assertEqual(3, lo)
        self.assertEqual(3, hi)

        lo, hi = acl.get_icmp_low_high("any")
        self.assertEqual(0, lo)
        self.assertEqual(255, hi)

        lo, hi = acl.get_icmp_low_high("10-")
        self.assertEqual(10, lo)
        self.assertEqual(255, hi)

        lo, hi = acl.get_icmp_low_high("-10")
        self.assertEqual(0, lo)
        self.assertEqual(10, hi)

        lo, hi = acl.get_icmp_low_high("10-20")
        self.assertEqual(10, lo)
        self.assertEqual(20, hi)

    def test_is_ip(self):
        self.assertTrue(acl.is_ip("192.0.2.1"))
        self.assertTrue(acl.is_ip("192.0.2.1/24"))
        self.assertTrue(acl.is_ip("192.0.2.0/24"))
        self.assertTrue(acl.is_ip("2001:db8::1"))
        self.assertTrue(acl.is_ip("2001:db8::1/64"))
        self.assertTrue(acl.is_ip("2001:db8::/64"))
        self.assertFalse(acl.is_ip(True))
        self.assertFalse(acl.is_ip("String"))
        self.assertFalse(acl.is_ip([]))
        self.assertFalse(acl.is_ip({}))

    def test_get_network_list(self):
        for s in ["192.0.2.1", "192.0.2.1/24", "2001:db8::1", "2001:db8::1/64"]:
            l = acl.get_network_list(self.cfg, s)
            self.assertIsInstance(l, list)
            self.assertEqual(1, len(l))
            n = l[0]

        l = acl.get_network_list(self.cfg, "trusted")
        self.assertIsInstance(l, list)
        self.assertEqual(5, len(l))

        l = acl.get_network_list(self.cfg, "trusted", want_ipv6=False)
        self.assertIsInstance(l, list)
        self.assertEqual(2, len(l))

        l = acl.get_network_list(self.cfg, "trusted", want_ipv4=False)
        self.assertIsInstance(l, list)
        self.assertEqual(3, len(l))

        l = acl.get_network_list(self.cfg, "trusted", want_ipv4=False, want_ipv6=False)
        self.assertIsInstance(l, list)
        self.assertEqual(0, len(l))

        l = acl.get_network_list(self.cfg, "pl-notexist")
        self.assertIsInstance(l, list)
        self.assertEqual(0, len(l))

    def test_network_list_has_family(self):
        l = acl.get_network_list(self.cfg, "trusted")
        self.assertTrue(acl.network_list_has_family(l, 4))
        self.assertTrue(acl.network_list_has_family(l, 6))

        l = acl.get_network_list(self.cfg, "trusted", want_ipv4=False)
        self.assertFalse(acl.network_list_has_family(l, 4))
        self.assertTrue(acl.network_list_has_family(l, 6))

        l = acl.get_network_list(self.cfg, "trusted", want_ipv6=False)
        self.assertTrue(acl.network_list_has_family(l, 4))
        self.assertFalse(acl.network_list_has_family(l, 6))

        l = acl.get_network_list(self.cfg, "trusted", want_ipv4=False, want_ipv6=False)
        self.assertFalse(acl.network_list_has_family(l, 4))
        self.assertFalse(acl.network_list_has_family(l, 6))



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A vppcfg configuration module that handles addresses"""
import ipaddress


def get_all_addresses_except_ifname(yaml, except_ifname):
    """Return a list of all ipaddress.ip_interface() instances in the entire config,
    except for those that belong to 'ifname'.
    """
    ret = []
    if "interfaces" in yaml:
        for ifname, iface in yaml["interfaces"].items():
            if ifname == except_ifname:
                continue

            if "addresses" in iface:
                for addr in iface["addresses"]:
                    ret.append(ipaddress.ip_interface(addr))
            if "sub-interfaces" in iface:
                for subid, sub_iface in iface["sub-interfaces"].items():
                    sub_ifname = f"{ifname}.{int(subid)}"
                    if sub_ifname == except_ifname:
                        continue

                    if "addresses" in sub_iface:
                        for addr in sub_iface["addresses"]:
                            ret.append(ipaddress.ip_interface(addr))
    if "loopbacks" in yaml:
        for ifname, iface in yaml["loopbacks"].items():
            if ifname == except_ifname:
                continue

            if "addresses" in iface:
                for addr in iface["addresses"]:
                    ret.append(ipaddress.ip_interface(addr))
    if "bridgedomains" in yaml:
        for ifname, iface in yaml["bridgedomains"].items():
            if ifname == except_ifname:
                continue

            if "addresses" in iface:
                for addr in iface["addresses"]:
                    ret.append(ipaddress.ip_interface(addr))

    return ret


def is_allowed(yaml, ifname, iface_addresses, ip_interface):
    """Returns True if there is at most one occurence of the ip_interface (an IPv4/IPv6 prefix+len)
    in the entire config. That said, we need the 'iface_addresses' because VPP is a bit fickle in
    this regard.

    IP addresses from the same prefix/len can be added to a given interface (ie 192.0.2.1/24 and
    192.0.2.2/24), but other than that, any prefix can not occur as a more-specific or less-specific
    of any other interface.

    So, we will allow:
    - any ip_interface that is of equal network/len of existing one(s) _on the same interface_

    And, we will reject
    - any ip_interface that is a more specific of any existing one
    - any ip_interface that is a less specific of any existing one

    Examples:
    vpp# set interface ip address loop0 192.0.2.1/24
    vpp# set interface ip address loop0 192.0.2.2/24
    vpp# set interface ip address loop0 192.0.2.1/29
    set interface ip address: failed to add 192.0.2.1/29 on loop0 which conflicts with 192.0.2.1/24 for interface loop0
    vpp# set interface ip address loop0 192.0.2.3/23
    set interface ip address: failed to add 192.0.2.3/23 on loop0 which conflicts with 192.0.2.1/24 for interface loop0
    """
    all_other_addresses = get_all_addresses_except_ifname(yaml, ifname)

    my_ip_network = ipaddress.ip_network(ip_interface, strict=False)

    for ipi in all_other_addresses:
        if ipi.version != my_ip_network.version:
            continue

        if ipaddress.ip_network(ipi, strict=False) == my_ip_network:
            return False

        if ipaddress.ip_network(ipi, strict=False).subnet_of(my_ip_network):
            return False

        if my_ip_network.subnet_of(ipaddress.ip_network(ipi, strict=False)):
            return False

    for addr in iface_addresses:
        ipi = ipaddress.ip_interface(addr)
        if ipi.version != my_ip_network.version:
            continue

        if ipaddress.ip_network(ipi, strict=False) == my_ip_network:
            return True

        if ipaddress.ip_network(ipi, strict=False).subnet_of(my_ip_network):
            return False

        if my_ip_network.subnet_of(ipaddress.ip_network(ipi, strict=False)):
            return False

    return True


def get_canonical(iface_address):
    """Returns the canonical form of an interface address, which can be either an address
    '2001:db8::1' or a prefix '2001:db8::1/64' for either IPv4 of IPv6.

    This function the string representation of the canonical address."""

    is_prefix = "/" in iface_address
    if is_prefix:
        iface_address, prefixlen = iface_address.split("/")

    ip_address = ipaddress.ip_address(iface_address)
    if is_prefix:
        return str(ip_address) + "/" + prefixlen
    return str(ip_address)


def is_canonical(iface_address):
    """Checks to see that the interface address is written in a canonical form, useful to
    ensure that IPv6 addresses are written in such a way that they don't trigger spurious
    diffs when comparing to VPP. As an example, '2001:DB8:0:0::1/64' is fine, but VPP will
    show this as '2001:db8::1/64'.

    Input can be either an address 2001:db8::1 or a prefix 2001:db8::1/128

    This function returns False if the iface_address isn't canonical, and True otherwise.
    """
    can = get_canonical(iface_address)
    return can == iface_address



#
# Copyright (c) 2023 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A vppcfg configuration module that validates acls"""
import logging
import socket
import ipaddress
from . import prefixlist


def get_acls(yaml):
    """Return a list of all acls."""
    ret = []
    if "acls" in yaml:
        for aclname, _acl in yaml["acls"].items():
            ret.append(aclname)
    return ret


def get_by_name(yaml, aclname):
    """Return the acl by name, if it exists. Return None otherwise."""
    try:
        if aclname in yaml["acls"]:
            return aclname, yaml["acls"][aclname]
    except KeyError:
        pass
    return None, None


def hydrate_term(acl_term):
    """Adds all defaults to an ACL term"""

    if "family" not in acl_term:
        acl_term["family"] = "any"
    if "source" not in acl_term:
        acl_term["source"] = "any"
    if "destination" not in acl_term:
        acl_term["destination"] = "any"

    if "protocol" not in acl_term or acl_term["protocol"] == "any":
        acl_term["protocol"] = 0

    if "source-port" not in acl_term:
        acl_term["source-port"] = "any"
    if "destination-port" not in acl_term:
        acl_term["destination-port"] = "any"
    if "icmp-code" not in acl_term:
        acl_term["icmp-code"] = "any"
    if "icmp-type" not in acl_term:
        acl_term["icmp-type"] = "any"

    return acl_term


def get_icmp_low_high(icmpstring):
    """For a given icmp string, which can be either an integer or a range of
    integers including start/stop being omitted, eg 0-255, 10- or -10, or the
    string "any", return a tuple of (lowport, highport) or (None, None) upon
    error"""
    if isinstance(icmpstring, int):
        return int(icmpstring), int(icmpstring)
    if "any" == icmpstring:
        return 0, 255

    try:
        icmp = int(icmpstring)
        if icmp > 0:
            return icmp, icmp
    except ValueError:
        pass

    if icmpstring.startswith("-"):
        icmp = int(icmpstring[1:])
        return 0, icmp

    if icmpstring.endswith("-"):
        icmp = int(icmpstring[:-1])
        return icmp, 255

    try:
        icmps = icmpstring.split("-")
        return int(icmps[0]), int(icmps[1])
    except ValueError:
        pass

    return None, None


def get_port_low_high(portstring):
    """For a given port string, which can be either an integer, a symbolic port name
    in /etc/services, a range of integers including start/stop being omitted, eg
    0-65535, 1024- or -1024, or the string "any", return a tuple of
    (lowport, highport) or (None, None) upon error"""
    if isinstance(portstring, int):
        return int(portstring), int(portstring)
    if "any" == portstring:
        return 0, 65535

    try:
        port = int(portstring)
        if port > 0:
            return port, port
    except ValueError:
        pass

    try:
        port = socket.getservbyname(portstring)
        return port, port
    except OSError:
        pass

    if portstring.startswith("-"):
        port = int(portstring[1:])
        return 0, port

    if portstring.endswith("-"):
        port = int(portstring[:-1])
        return port, 65535

    try:
        ports = portstring.split("-")
        return int(ports[0]), int(ports[1])
    except ValueError:
        pass

    return None, None


def is_ip(ip_string):
    """Return True if the given ip_string is either an IPv4/IPv6 address or prefix."""
    if not isinstance(ip_string, str):
        return False

    try:
        _ = ipaddress.ip_network(ip_string, strict=False)
        return True
    except ValueError:
        pass
    return False


def get_network_list(yaml, network_string, want_ipv4=True, want_ipv6=True):
    """Return the full list of source or destination address(es). This function resolves the
    'source' or 'destination' field, which can either be an IP address, a Prefix, or the name
    of a Prefix List. It returns a list of ip_network() objects, including prefix. IP addresses
    will receive prefixlen /32 or /128. Optionally, want_ipv4 or want_ipv6 can be set to False
    to filter the list."""

    ret = []
    if is_ip(network_string):
        ipn = ipaddress.ip_network(network_string, strict=False)
        if ipn.version == 4 and want_ipv4:
            ret = [ipn]
        if ipn.version == 6 and want_ipv6:
            ret = [ipn]
        return ret

    if network_string == "any":
        if want_ipv4:
            ret.append(ipaddress.ip_network("0.0.0.0/0"))
        if want_ipv6:
            ret.append(ipaddress.ip_network("::/0"))
        return ret

    return prefixlist.get_network_list(
        yaml, network_string, want_ipv4=want_ipv4, want_ipv6=want_ipv6
    )


def get_protocol(protostring):
    """For a given protocol string, which can be either an integer or a symbolic port
    name in /etc/protocols, return the protocol number as integer, or None if it cannot
    be determined."""
    if isinstance(protostring, int):
        return int(protostring)
    if "any" == protostring:
        return 0

    try:
        proto = int(protostring)
        if proto > 0:
            return proto
    except ValueError:
        pass

    try:
        proto = socket.getprotobyname(protostring)
        return proto
    except OSError:
        pass

    return None


def network_list_has_family(network_list, version):
    """Returns True if the given list of ip_network() elements has at least one
    element with the specified version, which can be either 4 or 6. Return False
    otherwise"""
    for m in network_list:
        if m.version == version:
            return True
    return False


def validate_acls(yaml):
    """Validate the semantics of all YAML 'acls' entries"""
    result = True
    msgs = []
    logger = logging.getLogger("vppcfg.config")
    logger.addHandler(logging.NullHandler())

    if not "acls" in yaml:
        return result, msgs

    for aclname, acl in yaml["acls"].items():
        terms = 0
        for acl_term in acl["terms"]:
            terms += 1
            orig_acl_term = acl_term.copy()
            acl_term = hydrate_term(acl_term)
            logger.debug(
                f"acl {aclname} term {terms} orig {orig_acl_term} hydrated {acl_term}"
            )
            if acl_term["family"] == "ipv4":
                want_ipv4 = True
                want_ipv6 = False
            elif acl_term["family"] == "ipv6":
                want_ipv4 = False
                want_ipv6 = True
            else:
                want_ipv4 = True
                want_ipv6 = True

            src_network_list = get_network_list(
                yaml, acl_term["source"], want_ipv4=want_ipv4, want_ipv6=want_ipv6
            )
            dst_network_list = get_network_list(
                yaml, acl_term["destination"], want_ipv4=want_ipv4, want_ipv6=want_ipv6
            )
            logger.debug(
                f"acl {aclname} term {terms} src: {src_network_list} dst: {dst_network_list}"
            )
            if len(src_network_list) == 0:
                msgs.append(
                    f"acl {aclname} term {terms} family {acl_term['family']} has no source"
                )
                result = False
            if len(dst_network_list) == 0:
                msgs.append(
                    f"acl {aclname} term {terms} family {acl_term['family']} has no destination"
                )
                result = False
            if len(dst_network_list) == 0 or len(src_network_list) == 0:
                ## Pointless to continue if there's no src/dst at all
                continue

            src_network_has_ipv4 = network_list_has_family(src_network_list, 4)
            dst_network_has_ipv4 = network_list_has_family(dst_network_list, 4)
            src_network_has_ipv6 = network_list_has_family(src_network_list, 6)
            dst_network_has_ipv6 = network_list_has_family(dst_network_list, 6)

            if (
                src_network_has_ipv4 != dst_network_has_ipv4
                and src_network_has_ipv6 != dst_network_has_ipv6
            ):
                msgs.append(
                    f"acl {aclname} term {terms} source and destination family do not overlap"
                )
                result = False
                continue

            proto = get_protocol(acl_term["protocol"])
            if proto is None:
                msgs.append(f"acl {aclname} term {terms} could not understand protocol")
                result = False

            if not proto in [6, 17]:
                if "source-port" in orig_acl_term:
                    msgs.append(
                        f"acl {aclname} term {terms} source-port can only be specified for protocol tcp or udp"
                    )
                    result = False
                if "destination-port" in orig_acl_term:
                    msgs.append(
                        f"acl {aclname} term {terms} destination-port can only be specified for protocol tcp or udp"
                    )
                    result = False
            else:
                src_low_port, src_high_port = get_port_low_high(acl_term["source-port"])
                dst_low_port, dst_high_port = get_port_low_high(
                    acl_term["destination-port"]
                )

                if src_low_port is None or src_high_port is None:
                    msgs.append(
                        f"acl {aclname} term {terms} could not understand source-port"
                    )
                    result = False
                else:
                    if src_low_port > src_high_port:
                        msgs.append(
                            f"acl {aclname} term {terms} source-port low value is greater than high value"
                        )
                        result = False
                    if src_low_port < 0 or src_low_port > 65535:
                        msgs.append(
                            f"acl {aclname} term {terms} source-port low value is not between [0,65535]"
                        )
                        result = False
                    if src_high_port < 0 or src_high_port > 65535:
                        msgs.append(
                            f"acl {aclname} term {terms} source-port high value is not between [0,65535]"
                        )
                        result = False

                if dst_low_port is None or dst_high_port is None:
                    msgs.append(
                        f"acl {aclname} term {terms} could not understand destination-port"
                    )
                    result = False
                else:
                    if dst_low_port > dst_high_port:
                        msgs.append(
                            f"acl {aclname} term {terms} destination-port low value is greater than high value"
                        )
                        result = False
                    if dst_low_port < 0 or dst_low_port > 65535:
                        msgs.append(
                            f"acl {aclname} term {terms} destination-port low value is not between [0,65535]"
                        )
                        result = False
                    if dst_high_port < 0 or dst_high_port > 65535:
                        msgs.append(
                            f"acl {aclname} term {terms} destination-port high value is not between [0,65535]"
                        )
                        result = False

            if not proto in [1, 58]:
                if "icmp-code" in orig_acl_term:
                    msgs.append(
                        f"acl {aclname} term {terms} icmp-code can only be specified for protocol icmp or ipv6-icmp"
                    )
                    result = False
                if "icmp-type" in orig_acl_term:
                    msgs.append(
                        f"acl {aclname} term {terms} icmp-type can only be specified for protocol icmp or ipv6-icmp"
                    )
                    result = False
            else:
                icmp_code_low, icmp_code_high = get_icmp_low_high(acl_term["icmp-code"])
                icmp_type_low, icmp_type_high = get_icmp_low_high(acl_term["icmp-type"])
                if icmp_code_low > icmp_code_high:
                    msgs.append(
                        f"acl {aclname} term {terms} icmp-code low value is greater than high value"
                    )
                    result = False
                if icmp_type_low > icmp_type_high:
                    msgs.append(
                        f"acl {aclname} term {terms} icmp-type low value is greater than high value"
                    )
                    result = False

    return result, msgs



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A vppcfg configuration module that validates MAC addresses"""
import netaddr


def is_valid(mac):
    """Return True if the string given in `mac` is a valid (6-byte) MAC address,
    as defined by netaddr.EUI"""
    try:
        _addr = netaddr.EUI(mac)
    except netaddr.core.AddrFormatError:
        return False
    return True


def is_local(mac):
    """Return True if a MAC address is a valid locally administered one."""
    try:
        addr = netaddr.EUI(mac)
    except netaddr.core.AddrFormatError:
        return False
    return bool(addr.words[0] & 0b10)


def is_multicast(mac):
    """Return True if a MAC address is a valid multicast one."""
    try:
        addr = netaddr.EUI(mac)
    except netaddr.core.AddrFormatError:
        return False
    return bool(addr.words[0] & 0b01)


def is_unicast(mac):
    """Return True if a MAC address is a valid unicast one."""
    try:
        addr = netaddr.EUI(mac)
    except netaddr.core.AddrFormatError:
        return False
    return not bool(addr.words[0] & 0b01)



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""Unit tests for loopbacks"""
import unittest
import yaml
from . import loopback
from .unittestyaml import UnitTestYaml


class TestLoopbackMethods(unittest.TestCase):
    def setUp(self):
        with UnitTestYaml("test_loopback.yaml") as f:
            self.cfg = yaml.load(f, Loader=yaml.FullLoader)

    def test_get_by_lcp_name(self):
        ifname, iface = loopback.get_by_lcp_name(self.cfg, "loop56789012345")
        self.assertIsNotNone(iface)
        self.assertEqual("loop1", ifname)

        ifname, iface = loopback.get_by_lcp_name(self.cfg, "lcp-noexist")
        self.assertIsNone(iface)
        self.assertIsNone(ifname)

    def test_get_by_name(self):
        ifname, iface = loopback.get_by_name(self.cfg, "loop1")
        self.assertIsNotNone(iface)
        self.assertEqual("loop1", ifname)
        self.assertEqual(iface["mtu"], 2000)

        ifname, iface = loopback.get_by_name(self.cfg, "loop-noexist")
        self.assertIsNone(ifname)
        self.assertIsNone(iface)

    def test_enumerators(self):
        ifs = loopback.get_loopbacks(self.cfg)
        self.assertEqual(len(ifs), 3)
        self.assertIn("loop0", ifs)
        self.assertIn("loop1", ifs)
        self.assertIn("loop2", ifs)
        self.assertNotIn("loop-noexist", ifs)

    def test_is_mpls(self):
        self.assertTrue(loopback.is_mpls(self.cfg, "loop1"))
        self.assertFalse(loopback.is_mpls(self.cfg, "loop2"))
        self.assertFalse(loopback.is_mpls(self.cfg, "loop-noexist"))



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""Unit tests for bondethernet"""
import unittest
import yaml
from . import bondethernet
from .unittestyaml import UnitTestYaml


class TestBondEthernetMethods(unittest.TestCase):
    def setUp(self):
        with UnitTestYaml("test_bondethernet.yaml") as f:
            self.cfg = yaml.load(f, Loader=yaml.FullLoader)

    def test_get_by_name(self):
        ifname, iface = bondethernet.get_by_name(self.cfg, "BondEthernet0")
        self.assertIsNotNone(iface)
        self.assertEqual("BondEthernet0", ifname)
        self.assertIn("GigabitEthernet1/0/0", iface["interfaces"])
        self.assertNotIn("GigabitEthernet2/0/0", iface["interfaces"])

        ifname, iface = bondethernet.get_by_name(self.cfg, "BondEthernet-notexist")
        self.assertIsNone(iface)
        self.assertIsNone(ifname)

    def test_members(self):
        self.assertTrue(bondethernet.is_bond_member(self.cfg, "GigabitEthernet1/0/0"))
        self.assertTrue(bondethernet.is_bond_member(self.cfg, "GigabitEthernet1/0/1"))
        self.assertFalse(bondethernet.is_bond_member(self.cfg, "GigabitEthernet2/0/0"))
        self.assertFalse(
            bondethernet.is_bond_member(self.cfg, "GigabitEthernet2/0/0.100")
        )

    def test_is_bondethernet(self):
        self.assertTrue(bondethernet.is_bondethernet(self.cfg, "BondEthernet0"))
        self.assertFalse(
            bondethernet.is_bondethernet(self.cfg, "BondEthernet-notexist")
        )
        self.assertFalse(bondethernet.is_bondethernet(self.cfg, "GigabitEthernet1/0/0"))

    def test_enumerators(self):
        ifs = bondethernet.get_bondethernets(self.cfg)
        self.assertEqual(len(ifs), 3)
        self.assertIn("BondEthernet0", ifs)
        self.assertIn("BondEthernet1", ifs)
        self.assertIn("BondEthernet2", ifs)
        self.assertNotIn("BondEthernet-noexist", ifs)

    def test_get_mode(self):
        self.assertEqual("lacp", bondethernet.get_mode(self.cfg, "BondEthernet0"))
        self.assertEqual("xor", bondethernet.get_mode(self.cfg, "BondEthernet1"))

    def test_mode_to_int(self):
        self.assertEqual(1, bondethernet.mode_to_int("round-robin"))
        self.assertEqual(2, bondethernet.mode_to_int("active-backup"))
        self.assertEqual(3, bondethernet.mode_to_int("xor"))
        self.assertEqual(4, bondethernet.mode_to_int("broadcast"))
        self.assertEqual(5, bondethernet.mode_to_int("lacp"))
        self.assertEqual(-1, bondethernet.mode_to_int("not-exist"))

    def test_int_to_mode(self):
        self.assertEqual("round-robin", bondethernet.int_to_mode(1))
        self.assertEqual("active-backup", bondethernet.int_to_mode(2))
        self.assertEqual("xor", bondethernet.int_to_mode(3))
        self.assertEqual("broadcast", bondethernet.int_to_mode(4))
        self.assertEqual("lacp", bondethernet.int_to_mode(5))
        self.assertEqual("", bondethernet.int_to_mode(0))
        self.assertEqual("", bondethernet.int_to_mode(6))

    def test_get_lb(self):
        self.assertEqual("l34", bondethernet.get_lb(self.cfg, "BondEthernet0"))
        self.assertEqual("l2", bondethernet.get_lb(self.cfg, "BondEthernet1"))
        self.assertIsNone(bondethernet.get_lb(self.cfg, "BondEthernet2"))

    def test_lb_to_int(self):
        self.assertEqual(0, bondethernet.lb_to_int("l2"))
        self.assertEqual(1, bondethernet.lb_to_int("l34"))
        self.assertEqual(2, bondethernet.lb_to_int("l23"))
        self.assertEqual(3, bondethernet.lb_to_int("round-robin"))
        self.assertEqual(4, bondethernet.lb_to_int("broadcast"))
        self.assertEqual(5, bondethernet.lb_to_int("active-backup"))
        self.assertEqual(-1, bondethernet.lb_to_int("not-exist"))

    def test_int_to_lb(self):
        self.assertEqual("l2", bondethernet.int_to_lb(0))
        self.assertEqual("l34", bondethernet.int_to_lb(1))
        self.assertEqual("l23", bondethernet.int_to_lb(2))
        self.assertEqual("round-robin", bondethernet.int_to_lb(3))
        self.assertEqual("broadcast", bondethernet.int_to_lb(4))
        self.assertEqual("active-backup", bondethernet.int_to_lb(5))
        self.assertEqual("", bondethernet.int_to_lb(-1))



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""Unit tests for LCPs"""
import unittest
import yaml
from . import lcp
from . import interface
from .unittestyaml import UnitTestYaml


class TestLCPMethods(unittest.TestCase):
    def setUp(self):
        with UnitTestYaml("test_lcp.yaml") as f:
            self.cfg = yaml.load(f, Loader=yaml.FullLoader)

    def test_enumerators(self):
        lcps = lcp.get_lcps(self.cfg)
        self.assertIn("e1", lcps)
        self.assertIn("foo", lcps)
        self.assertIn("e2", lcps)
        loopback_lcps = lcp.get_lcps(self.cfg, interfaces=False, bridgedomains=False)
        self.assertIn("thrice", loopback_lcps)
        self.assertNotIn("e1", loopback_lcps)

    def test_lcp(self):
        self.assertTrue(lcp.is_unique(self.cfg, "e1"))
        self.assertTrue(lcp.is_unique(self.cfg, "foo"))
        self.assertTrue(lcp.is_unique(self.cfg, "notexist"))

        self.assertFalse(lcp.is_unique(self.cfg, "twice"))
        self.assertFalse(lcp.is_unique(self.cfg, "thrice"))

    def test_qinx(self):
        qinx_ifname, qinx_iface = interface.get_by_name(
            self.cfg, "GigabitEthernet1/0/1.201"
        )
        mid_ifname, mid_iface = interface.get_qinx_parent_by_name(
            self.cfg, "GigabitEthernet1/0/1.201"
        )
        parent_ifname, parent_iface = interface.get_parent_by_name(
            self.cfg, "GigabitEthernet1/0/1.201"
        )

        self.assertEqual(qinx_ifname, "GigabitEthernet1/0/1.201")
        self.assertEqual(mid_ifname, "GigabitEthernet1/0/1.200")
        self.assertEqual(parent_ifname, "GigabitEthernet1/0/1")

        qinx_ifname, qinx_iface = interface.get_by_name(
            self.cfg, "GigabitEthernet1/0/1.201"
        )
        mid_ifname, mid_iface = interface.get_qinx_parent_by_name(
            self.cfg, "GigabitEthernet1/0/1.201"
        )
        parent_ifname, parent_iface = interface.get_parent_by_name(
            self.cfg, "GigabitEthernet1/0/1.201"
        )

        self.assertEqual(qinx_ifname, "GigabitEthernet1/0/1.201")
        self.assertEqual(mid_ifname, "GigabitEthernet1/0/1.200")
        self.assertEqual(parent_ifname, "GigabitEthernet1/0/1")

        ifname, iface = interface.get_qinx_parent_by_name(
            self.cfg, "GigabitEthernet1/0/1.100"
        )
        self.assertIsNone(ifname)
        self.assertIsNone(iface)



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""Unit tests for taps"""
import unittest
import yaml
from . import prefixlist
from .unittestyaml import UnitTestYaml


class TestACLMethods(unittest.TestCase):
    def setUp(self):
        with UnitTestYaml("test_prefixlist.yaml") as f:
            self.cfg = yaml.load(f, Loader=yaml.FullLoader)

    def test_get_prefixlists(self):
        plist = prefixlist.get_prefixlists(self.cfg)
        self.assertIsInstance(plist, list)
        self.assertEqual(5, len(plist))

    def test_get_by_name(self):
        plname, _pl = prefixlist.get_by_name(self.cfg, "trusted")
        self.assertIsNotNone(_pl)
        self.assertEqual("trusted", plname)

        plname, _pl = prefixlist.get_by_name(self.cfg, "pl-noexist")
        self.assertIsNone(plname)
        self.assertIsNone(_pl)

    def test_count(self):
        v4, v6 = prefixlist.count(self.cfg, "trusted")
        self.assertEqual(2, v4)
        self.assertEqual(3, v6)

        v4, v6 = prefixlist.count(self.cfg, "empty")
        self.assertEqual(0, v4)
        self.assertEqual(0, v6)

        v4, v6 = prefixlist.count(self.cfg, "pl-noexist")
        self.assertEqual(0, v4)
        self.assertEqual(0, v6)

    def test_count_ipv4(self):
        self.assertEqual(2, prefixlist.count_ipv4(self.cfg, "trusted"))
        self.assertEqual(0, prefixlist.count_ipv4(self.cfg, "empty"))
        self.assertEqual(0, prefixlist.count_ipv4(self.cfg, "pl-noexist"))

    def test_count_ipv6(self):
        self.assertEqual(3, prefixlist.count_ipv6(self.cfg, "trusted"))
        self.assertEqual(0, prefixlist.count_ipv6(self.cfg, "empty"))
        self.assertEqual(0, prefixlist.count_ipv6(self.cfg, "pl-noexist"))

    def test_has_ipv4(self):
        self.assertTrue(prefixlist.has_ipv4(self.cfg, "trusted"))
        self.assertFalse(prefixlist.has_ipv4(self.cfg, "empty"))
        self.assertFalse(prefixlist.has_ipv4(self.cfg, "pl-noexist"))

    def test_has_ipv6(self):
        self.assertTrue(prefixlist.has_ipv6(self.cfg, "trusted"))
        self.assertFalse(prefixlist.has_ipv6(self.cfg, "empty"))
        self.assertFalse(prefixlist.has_ipv6(self.cfg, "pl-noexist"))

    def test_is_empty(self):
        self.assertFalse(prefixlist.is_empty(self.cfg, "trusted"))
        self.assertTrue(prefixlist.is_empty(self.cfg, "empty"))
        self.assertTrue(prefixlist.is_empty(self.cfg, "pl-noexist"))

    def test_get_network_list(self):
        l = prefixlist.get_network_list(self.cfg, "trusted")
        self.assertIsInstance(l, list)
        self.assertEqual(5, len(l))

        l = prefixlist.get_network_list(self.cfg, "trusted", want_ipv6=False)
        self.assertIsInstance(l, list)
        self.assertEqual(2, len(l))

        l = prefixlist.get_network_list(self.cfg, "trusted", want_ipv4=False)
        self.assertIsInstance(l, list)
        self.assertEqual(3, len(l))

        l = prefixlist.get_network_list(
            self.cfg, "trusted", want_ipv4=False, want_ipv6=False
        )
        self.assertIsInstance(l, list)
        self.assertEqual(0, len(l))

        l = prefixlist.get_network_list(self.cfg, "pl-notexist")
        self.assertIsInstance(l, list)
        self.assertEqual(0, len(l))



# Copyright (c) 2025 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""
Planner sync operations - handles synchronization of VPP dataplane configuration.
"""

from vppcfg.config import loopback
from vppcfg.config import interface
from vppcfg.config import bondethernet
from vppcfg.config import bridgedomain
from .planner_base import PlannerBase


class PlannerSyncOperations(PlannerBase):  # pylint: disable=too-few-public-methods
    """Mixin class providing sync operations for the Planner."""

    def sync(self):
        """Synchronize the VPP Dataplane configuration for all objects in the config"""
        ret = True
        if not self._sync_loopbacks():
            self.logger.warning("Could not sync Loopbacks in VPP")
            ret = False
        if not self._sync_bondethernets():
            self.logger.warning("Could not sync bondethernets in VPP")
            ret = False
        if not self._sync_bridgedomains():
            self.logger.warning("Could not sync bridgedomains in VPP")
            ret = False
        if not self._sync_l2xcs():
            self.logger.warning("Could not sync L2 Cross Connects in VPP")
            ret = False
        if not self._sync_mtu():
            self.logger.warning("Could not sync interface MTU in VPP")
            ret = False
        if not self._sync_addresses():
            self.logger.warning("Could not sync interface addresses in VPP")
            ret = False
        if not self._sync_unnumbered():
            self.logger.warning("Could not sync unnumbered interfaces in VPP")
            ret = False
        if not self._sync_phys():
            self.logger.warning("Could not sync PHYs in VPP")
            ret = False
        if not self._sync_mpls_state():
            self.logger.warning("Could not sync interface MPLS state in VPP")
            ret = False
        if not self._sync_sflow_state():
            self.logger.warning("Could not sync interface sFlow state in VPP")
            ret = False
        if not self._sync_admin_state():
            self.logger.warning("Could not sync interface adminstate in VPP")
            ret = False
        return ret

    def _sync_loopbacks(self):
        """Synchronize the VPP Dataplane configuration for loopbacks"""
        for ifname in loopback.get_loopbacks(self.cfg):
            if not ifname in self.vpp.cache["interface_names"]:
                ## New loopback
                continue
            vpp_iface = self.vpp.get_interface_by_name(ifname)
            if not vpp_iface:
                continue

            config_ifname, config_iface = loopback.get_by_name(self.cfg, ifname)
            if "mac" in config_iface and config_iface["mac"] != str(
                vpp_iface.l2_address
            ):
                cli = f"set interface mac address {config_ifname} {config_iface['mac']}"
                self.cli["sync"].append(cli)
        return True

    def _sync_phys(self):
        """Synchronize the VPP Dataplane configuration for PHYs"""
        for ifname in interface.get_phys(self.cfg):
            if not ifname in self.vpp.cache["interface_names"]:
                ## New interface
                continue
            vpp_iface = self.vpp.get_interface_by_name(ifname)
            if not vpp_iface:
                continue

            config_ifname, config_iface = interface.get_by_name(self.cfg, ifname)
            if "mac" in config_iface and config_iface["mac"] != str(
                vpp_iface.l2_address
            ):
                cli = f"set interface mac address {config_ifname} {config_iface['mac']}"
                self.cli["sync"].append(cli)
        return True

    def _sync_bondethernets(self):
        """Synchronize the VPP Dataplane configuration for bondethernets"""
        for ifname in bondethernet.get_bondethernets(self.cfg):
            vpp_iface = self.vpp.get_interface_by_name(ifname)
            if vpp_iface:
                vpp_members = [
                    self.vpp.cache["interfaces"][x].interface_name
                    for x in self.vpp.cache["bondethernet_members"][
                        vpp_iface.sw_if_index
                    ]
                ]
            else:
                ## New BondEthernet
                vpp_members = []

            config_bond_ifname, config_bond_iface = bondethernet.get_by_name(
                self.cfg, ifname
            )
            if not "interfaces" in config_bond_iface:
                continue
            config_ifname, config_iface = interface.get_by_name(self.cfg, ifname)
            bondmac = None
            for member_ifname in sorted(config_bond_iface["interfaces"]):
                member_ifname, member_iface = interface.get_by_name(
                    self.cfg, member_ifname
                )
                member_iface = self.vpp.get_interface_by_name(member_ifname)
                if not member_iface or member_ifname not in vpp_members:
                    if (
                        len(vpp_members) == 0
                        and member_iface
                        and member_iface.l2_address != "00:00:00:00:00:00"
                    ):
                        bondmac = member_iface.l2_address
                    cli = f"bond add {config_bond_ifname} {member_ifname}"
                    self.cli["sync"].append(cli)
            if (
                vpp_iface
                and "mac" in config_iface
                and str(vpp_iface.l2_address) != config_iface["mac"]
            ):
                cli = f"set interface mac address {config_ifname} {config_iface['mac']}"
                self.cli["sync"].append(cli)
            elif bondmac and "lcp" in config_iface:
                ## TODO(pim) - Ensure LCP has the same MAC as the BondEthernet
                ## VPP, when creating a BondEthernet, will give it an ephemeral MAC. Then, when the
                ## first member is enslaved, the MAC address changes to that of the first member.
                ## However, LinuxCP does not propagate this change to the Linux side (because there
                ## is no API callback for MAC address changes). To ensure consistency, every time we
                ## sync members, we ought to ensure the Linux device has the same MAC as its BondEthernet.
                cli = (
                    f"comment {{ ip link set {config_iface['lcp']} address {bondmac} }}"
                )
                self.cli["sync"].append(cli)
        return True

    def _sync_bridgedomains(self):
        """Synchronize the VPP Dataplane configuration for bridgedomains"""
        for ifname in bridgedomain.get_bridgedomains(self.cfg):
            instance = int(ifname[2:])
            if instance in self.vpp.cache["bridgedomains"]:
                vpp_bridge = self.vpp.cache["bridgedomains"][instance]
                bvi_sw_if_index = vpp_bridge.bvi_sw_if_index
                bridge_sw_if_index_list = [
                    x.sw_if_index for x in vpp_bridge.sw_if_details
                ]
                bridge_members = [
                    self.vpp.cache["interfaces"][x].interface_name
                    for x in bridge_sw_if_index_list
                    if x in self.vpp.cache["interfaces"]
                ]
            else:
                ## New BridgeDomain
                vpp_bridge = None
                bvi_sw_if_index = -1
                bridge_members = []

            config_bridge_ifname, config_bridge_iface = bridgedomain.get_by_name(
                self.cfg, f"bd{int(instance)}"
            )
            if vpp_bridge:
                # Sync settings on existing bridge. _create_bridgedomain() will have set them for new bridges.
                settings = bridgedomain.get_settings(self.cfg, config_bridge_ifname)
                if settings["learn"] != vpp_bridge.learn:
                    cli = f"set bridge-domain learn {int(instance)}"
                    if not settings["learn"]:
                        cli += " disable"
                    self.cli["sync"].append(cli)
                if settings["unicast-forward"] != vpp_bridge.forward:
                    cli = f"set bridge-domain forward {int(instance)}"
                    if not settings["unicast-forward"]:
                        cli += " disable"
                    self.cli["sync"].append(cli)
                if settings["unicast-flood"] != vpp_bridge.flood:
                    cli = f"set bridge-domain flood {int(instance)}"
                    if not settings["unicast-flood"]:
                        cli += " disable"
                    self.cli["sync"].append(cli)
                if settings["unknown-unicast-flood"] != vpp_bridge.uu_flood:
                    cli = f"set bridge-domain uu-flood {int(instance)}"
                    if not settings["unknown-unicast-flood"]:
                        cli += " disable"
                    self.cli["sync"].append(cli)
                if settings["arp-termination"] != vpp_bridge.arp_term:
                    cli = f"set bridge-domain arp term {int(instance)}"
                    if not settings["arp-termination"]:
                        cli += " disable"
                    self.cli["sync"].append(cli)
                if settings["arp-unicast-forward"] != vpp_bridge.arp_ufwd:
                    cli = f"set bridge-domain arp-ufwd {int(instance)}"
                    if not settings["arp-unicast-forward"]:
                        cli += " disable"
                    self.cli["sync"].append(cli)
                if settings["mac-age-minutes"] != vpp_bridge.mac_age:
                    cli = f"set bridge-domain mac-age {int(instance)} {int(settings['mac-age-minutes'])}"
                    self.cli["sync"].append(cli)

            if "bvi" in config_bridge_iface:
                bviname = config_bridge_iface["bvi"]
                bvi_iface = self.vpp.get_interface_by_name(bviname)
                if not bvi_iface or bvi_iface.sw_if_index != bvi_sw_if_index:
                    cli = f"set interface l2 bridge {bviname} {int(instance)} bvi"
                    self.cli["sync"].append(cli)

            if "interfaces" in config_bridge_iface:
                for member_ifname in config_bridge_iface["interfaces"]:
                    member_ifname, _member_iface = interface.get_by_name(
                        self.cfg, member_ifname
                    )
                    if not member_ifname in bridge_members:
                        cli = f"set interface l2 bridge {member_ifname} {int(instance)}"
                        self.cli["sync"].append(cli)
                        operation = "disable"
                        if interface.is_qinx(self.cfg, member_ifname):
                            operation = "pop 2"
                        elif interface.is_sub(self.cfg, member_ifname):
                            operation = "pop 1"
                        cli = (
                            f"set interface l2 tag-rewrite {member_ifname} {operation}"
                        )
                        self.cli["sync"].append(cli)
        return True

    def _sync_l2xcs(self):
        """Synchronize the VPP Dataplane configuration for L2 cross connects"""
        for ifname in interface.get_l2xc_interfaces(self.cfg):
            config_rx_ifname, config_rx_iface = interface.get_by_name(self.cfg, ifname)
            config_tx_ifname, _config_tx_iface = interface.get_by_name(
                self.cfg, config_rx_iface["l2xc"]
            )
            vpp_rx_iface = self.vpp.get_interface_by_name(config_rx_ifname)
            vpp_tx_iface = self.vpp.get_interface_by_name(config_tx_ifname)

            l2xc_changed = False
            if not vpp_rx_iface or not vpp_tx_iface:
                l2xc_changed = True
            elif not vpp_rx_iface.sw_if_index in self.vpp.cache["l2xcs"]:
                l2xc_changed = True
            elif (
                not vpp_tx_iface.sw_if_index
                == self.vpp.cache["l2xcs"][vpp_rx_iface.sw_if_index].tx_sw_if_index
            ):
                l2xc_changed = True

            if l2xc_changed:
                cli = f"set interface l2 xconnect {config_rx_ifname} {config_tx_ifname}"
                self.cli["sync"].append(cli)

                operation = "disable"
                if interface.is_qinx(self.cfg, config_rx_ifname):
                    operation = "pop 2"
                elif interface.is_sub(self.cfg, config_rx_ifname):
                    operation = "pop 1"
                cli = f"set interface l2 tag-rewrite {config_rx_ifname} {operation}"
                self.cli["sync"].append(cli)
        return True

    def _sync_mtu_direction(self, shrink=True):
        """Synchronize the VPP Dataplane packet MTU, where 'shrink' determines the
        direction (if shrink is True, go from inner-most (QinQ) to outer-most (untagged),
        and the other direction if shrink is False"""
        if shrink:
            tag_list = [2, 1, 0]
        else:
            tag_list = [0, 1, 2]

        for numtags in tag_list:
            for ifname in loopback.get_loopbacks(self.cfg) + interface.get_interfaces(
                self.cfg
            ):
                if numtags == 0 and interface.is_sub(self.cfg, ifname):
                    continue
                if numtags == 1 and not interface.is_sub(self.cfg, ifname):
                    continue
                if numtags == 1 and interface.is_qinx(self.cfg, ifname):
                    continue
                if numtags == 2 and not interface.is_qinx(self.cfg, ifname):
                    continue
                config_mtu = 1500
                vpp_mtu = 9000
                if ifname.startswith("loop"):
                    _iface = self.vpp.get_interface_by_name(ifname)
                    if _iface:
                        vpp_mtu = _iface.mtu[0]
                    vpp_ifname, config_iface = loopback.get_by_name(self.cfg, ifname)
                    if "mtu" in config_iface:
                        config_mtu = config_iface["mtu"]
                else:
                    if numtags > 0:
                        vpp_mtu = 0
                    _iface = self.vpp.get_interface_by_name(ifname)
                    if _iface:
                        vpp_mtu = _iface.mtu[0]
                    vpp_ifname, config_iface = interface.get_by_name(self.cfg, ifname)
                    config_mtu = interface.get_mtu(self.cfg, ifname)

                if shrink and config_mtu < vpp_mtu:
                    cli = f"set interface mtu packet {int(config_mtu)} {vpp_ifname}"
                    self.cli["sync"].append(cli)
                elif not shrink and config_mtu > vpp_mtu:
                    cli = f"set interface mtu packet {int(config_mtu)} {vpp_ifname}"
                    self.cli["sync"].append(cli)
        return True

    def _sync_link_mtu_direction(self, shrink=True):
        """Synchronize the VPP Dataplane max frame size (link MTU), where 'shrink' determines the
        direction (if shrink is True, go from inner-most (QinQ) to outer-most (untagged),
        and the other direction if shrink is False"""
        lcp_tap_sw_if_indexes = {lcp.host_sw_if_index for lcp in self.vpp.cache["lcps"].values()}
        for _idx, vpp_iface in self.vpp.cache["interfaces"].items():
            if vpp_iface.sub_number_of_tags != 0:
                continue
            if vpp_iface.interface_dev_type in ["local", "Loopback", "VXLAN", "virtio"]:
                continue
            if vpp_iface.sw_if_index in lcp_tap_sw_if_indexes:
                continue

            _config_ifname, config_iface = interface.get_by_name(
                self.cfg, vpp_iface.interface_name
            )
            if not config_iface:
                self.logger.warning(
                    f"Interface {vpp_iface.interface_name} exists in VPP but not in config, this is dangerous"
                )
                continue
            if not interface.is_phy(self.cfg, vpp_iface.interface_name):
                continue
            config_mtu = interface.get_mtu(self.cfg, vpp_iface.interface_name)

            if (
                vpp_iface.interface_dev_type == "bond"
                and vpp_iface.link_mtu < config_mtu
            ):
                self.logger.warning(
                    f"{vpp_iface.interface_name} has a Max Frame Size ({vpp_iface.link_mtu}) "
                    "lower than desired MTU ({config_mtu}), this is unsupported"
                )
                continue

            if shrink and config_mtu < vpp_iface.link_mtu:
                ## If the interface is up, temporarily down it in order to change the Max Frame Size
                if vpp_iface.flags & 1:  # IF_STATUS_API_FLAG_ADMIN_UP
                    cli = f"set interface state {vpp_iface.interface_name} down"
                    self.cli["sync"].append(cli)

                cli = f"set interface mtu {int(config_mtu)} {vpp_iface.interface_name}"
                self.cli["sync"].append(cli)

                if vpp_iface.flags & 1:  # IF_STATUS_API_FLAG_ADMIN_UP
                    cli = f"set interface state {vpp_iface.interface_name} up"
                    self.cli["sync"].append(cli)
            elif not shrink and config_mtu > vpp_iface.link_mtu:
                ## If the interface is up, temporarily down it in order to change the Max Frame Size
                if vpp_iface.flags & 1:  # IF_STATUS_API_FLAG_ADMIN_UP
                    cli = f"set interface state {vpp_iface.interface_name} down"
                    self.cli["sync"].append(cli)

                cli = f"set interface mtu {int(config_mtu)} {vpp_iface.interface_name}"
                self.cli["sync"].append(cli)

                if vpp_iface.flags & 1:  # IF_STATUS_API_FLAG_ADMIN_UP
                    cli = f"set interface state {vpp_iface.interface_name} up"
                    self.cli["sync"].append(cli)
        return True

    def _sync_mtu(self):
        """Synchronize the VPP Dataplane configuration for interface MTU"""
        ret = True
        if not self._sync_link_mtu_direction(shrink=False):
            self.logger.warning(
                "Could not sync growing interface Max Frame Size in VPP"
            )
            ret = False
        if not self._sync_link_mtu_direction(shrink=True):
            self.logger.warning(
                "Could not sync shrinking interface Max Frame Size in VPP"
            )
            ret = False
        if not self._sync_mtu_direction(shrink=True):
            self.logger.warning("Could not sync shrinking interface MTU in VPP")
            ret = False
        if not self._sync_mtu_direction(shrink=False):
            self.logger.warning("Could not sync growing interface MTU in VPP")
            ret = False
        return ret

    def _sync_sflow_state(self):
        """Synchronize the VPP Dataplane configuration and phy sFlow state"""

        if "sflow" in self.cfg and self.vpp.cache["sflow"]:
            if "header-bytes" in self.cfg["sflow"]:
                if (
                    self.vpp.cache["sflow"]["header-bytes"]
                    != self.cfg["sflow"]["header-bytes"]
                ):
                    cli = f"sflow header-bytes {self.cfg['sflow']['header-bytes']}"
                    self.cli["sync"].append(cli)
            if "polling-interval" in self.cfg["sflow"]:
                if (
                    self.vpp.cache["sflow"]["polling-interval"]
                    != self.cfg["sflow"]["polling-interval"]
                ):
                    cli = f"sflow polling-interval {self.cfg['sflow']['polling-interval']}"
                    self.cli["sync"].append(cli)
            if "sampling-rate" in self.cfg["sflow"]:
                if (
                    self.vpp.cache["sflow"]["sampling-rate"]
                    != self.cfg["sflow"]["sampling-rate"]
                ):
                    cli = f"sflow sampling-rate {self.cfg['sflow']['sampling-rate']}"
                    self.cli["sync"].append(cli)

        for ifname in interface.get_interfaces(self.cfg):
            vpp_ifname, config_iface = interface.get_by_name(self.cfg, ifname)

            try:
                config_sflow = config_iface["sflow"]
            except KeyError:
                config_sflow = False

            vpp_sflow = False
            if vpp_ifname in self.vpp.cache["interface_names"]:
                hw_if_index = self.vpp.cache["interface_names"][vpp_ifname]
                try:
                    vpp_sflow = self.vpp.cache["interface_sflow"][hw_if_index]
                except KeyError:
                    pass
            if vpp_sflow != config_sflow:
                if config_sflow:
                    cli = f"sflow enable {vpp_ifname}"
                else:
                    cli = f"sflow enable-disable {vpp_ifname} disable"
                self.cli["sync"].append(cli)
        return True

    def _sync_mpls_state(self):
        """Synchronize the VPP Dataplane configuration for interface and loopback MPLS state"""
        for ifname in loopback.get_loopbacks(self.cfg) + interface.get_interfaces(
            self.cfg
        ):
            if ifname.startswith("loop"):
                vpp_ifname, config_iface = loopback.get_by_name(self.cfg, ifname)
            else:
                vpp_ifname, config_iface = interface.get_by_name(self.cfg, ifname)

            try:
                config_mpls = config_iface["mpls"]
            except KeyError:
                config_mpls = False

            vpp_mpls = False
            if vpp_ifname in self.vpp.cache["interface_names"]:
                sw_if_index = self.vpp.cache["interface_names"][vpp_ifname]
                try:
                    vpp_mpls = self.vpp.cache["interface_mpls"][sw_if_index]
                except KeyError:
                    pass
            if vpp_mpls != config_mpls:
                state = "disable"
                if config_mpls:
                    state = "enable"
                cli = f"set interface mpls {vpp_ifname} {state}"
                self.cli["sync"].append(cli)
        return True

    def _sync_unnumbered(self):
        """Synchronize the VPP Dataplane configuration for unnumbered interface"""
        for ifname in loopback.get_loopbacks(self.cfg) + interface.get_interfaces(
            self.cfg
        ):
            if ifname.startswith("loop"):
                config_ifname, config_iface = loopback.get_by_name(self.cfg, ifname)
            else:
                config_ifname, config_iface = interface.get_by_name(self.cfg, ifname)

            config_unnumbered_ifname = None
            if "unnumbered" in config_iface:
                config_unnumbered_ifname = config_iface["unnumbered"]
            self.logger.debug(
                f"unnumbered iface {config_ifname} use {config_unnumbered_ifname}"
            )

            vpp_iface = self.vpp.get_interface_by_name(config_ifname)
            vpp_iface_unnumbered = self.vpp.get_interface_by_name(
                config_unnumbered_ifname
            )
            self.logger.debug(
                f"unnumbered iface {vpp_iface} use {vpp_iface_unnumbered}"
            )

            if not config_unnumbered_ifname:
                if (
                    vpp_iface
                    and vpp_iface.sw_if_index in self.vpp.cache["interface_unnumbered"]
                ):
                    cli = f"set interface unnumbered del {config_ifname}"
                    self.cli["sync"].append(cli)
                    del self.vpp.cache["interface_unnumbered"][vpp_iface.sw_if_index]
                    continue
                continue

            if (
                vpp_iface_unnumbered
                and vpp_iface
                and vpp_iface.sw_if_index in self.vpp.cache["interface_unnumbered"]
            ):
                if (
                    self.vpp.cache["interface_unnumbered"][vpp_iface.sw_if_index]
                    == vpp_iface_unnumbered.sw_if_index
                ):
                    continue

            cli = f"set interface unnumbered {config_ifname} use {config_unnumbered_ifname}"
            self.cli["sync"].append(cli)

        return True

    def _sync_addresses(self):
        """Synchronize the VPP Dataplane configuration for interface addresses"""
        for ifname in loopback.get_loopbacks(self.cfg) + interface.get_interfaces(
            self.cfg
        ):
            config_addresses = []
            vpp_addresses = []
            if ifname.startswith("loop"):
                vpp_ifname, config_iface = loopback.get_by_name(self.cfg, ifname)
                if "addresses" in config_iface:
                    config_addresses = config_iface["addresses"]
            else:
                vpp_ifname, config_iface = interface.get_by_name(self.cfg, ifname)
                if "addresses" in config_iface:
                    config_addresses = config_iface["addresses"]
            if vpp_ifname in self.vpp.cache["interface_names"]:
                _iface = self.vpp.get_interface_by_name(vpp_ifname)
                if _iface.sw_if_index in self.vpp.cache["interface_addresses"]:
                    vpp_addresses = [
                        str(x)
                        for x in self.vpp.cache["interface_addresses"][
                            _iface.sw_if_index
                        ]
                    ]
            for addr in config_addresses:
                if addr in vpp_addresses:
                    continue
                cli = f"set interface ip address {vpp_ifname} {addr}"
                self.cli["sync"].append(cli)
        return True

    def _sync_admin_state(self):
        """Synchronize the VPP Dataplane configuration for interface admin state"""
        for ifname in interface.get_interfaces(self.cfg) + loopback.get_loopbacks(
            self.cfg
        ):
            if ifname.startswith("loop"):
                vpp_ifname, _config_iface = loopback.get_by_name(self.cfg, ifname)
                config_admin_state = 1
            else:
                vpp_ifname, _config_iface = interface.get_by_name(self.cfg, ifname)
                config_admin_state = interface.get_admin_state(self.cfg, ifname)

            vpp_admin_state = 0
            _iface = self.vpp.get_interface_by_name(vpp_ifname)
            if _iface:
                vpp_admin_state = _iface.flags & 1  # IF_STATUS_API_FLAG_ADMIN_UP

            if config_admin_state == vpp_admin_state:
                continue
            state = "up"
            if config_admin_state == 0:
                state = "down"
            cli = f"set interface state {vpp_ifname} {state}"
            self.cli["sync"].append(cli)
        return True



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A vppcfg configuration module that validates taps"""
import logging
from . import mac


def get_taps(yaml):
    """Return a list of all taps."""
    ret = []
    if "taps" in yaml:
        for ifname, _iface in yaml["taps"].items():
            ret.append(ifname)
    return ret


def get_by_name(yaml, ifname):
    """Return the tap by name, if it exists. Return None otherwise."""
    try:
        if ifname in yaml["taps"]:
            return ifname, yaml["taps"][ifname]
    except KeyError:
        pass
    return None, None


def is_tap(yaml, ifname):
    """Returns True if the interface name is an existing tap in the config.
    The TAP has to be explicitly named in the configuration, and notably
    a TAP belonging to a Linux Control Plane (LCP) will return False.
    """
    ifname, iface = get_by_name(yaml, ifname)
    return iface is not None


def is_host_name_unique(yaml, hostname):
    """Returns True if there is at most one occurence of the given ifname amonst all host-names of TAPs."""
    if not "taps" in yaml:
        return True
    host_names = []
    for _tap_ifname, tap_iface in yaml["taps"].items():
        host_names.append(tap_iface["host"]["name"])
    return host_names.count(hostname) < 2


def validate_taps(yaml):
    """Validate the semantics of all YAML 'taps' entries"""
    result = True
    msgs = []
    logger = logging.getLogger("vppcfg.config")
    logger.addHandler(logging.NullHandler())

    if not "taps" in yaml:
        return result, msgs

    for ifname, iface in yaml["taps"].items():
        logger.debug(f"tap {iface}")
        instance = int(ifname[3:])

        ## NOTE(pim): 1024 is not off-by-one, tap1024 is precisely the highest permissible id
        if instance > 1024:
            msgs.append(f"tap {ifname} has instance {int(instance)} which is too large")
            result = False

        if not is_host_name_unique(yaml, iface["host"]["name"]):
            msgs.append(
                f"tap {ifname} does not have a unique host name {iface['host']['name']}"
            )
            result = False

        if "rx-ring-size" in iface:
            ncount = iface["rx-ring-size"]
            if ncount & (ncount - 1) != 0:
                msgs.append(f"tap {ifname} rx-ring-size must be a power of two")
                result = False

        if "tx-ring-size" in iface:
            ncount = iface["tx-ring-size"]
            if ncount & (ncount - 1) != 0:
                msgs.append(f"tap {ifname} tx-ring-size must be a power of two")
                result = False

        if (
            "namespace-create" in iface["host"]
            and iface["host"]["namespace-create"]
            and not "namespace" in iface["host"]
        ):
            msgs.append(
                f"tap {ifname} namespace-create can only be set if namespace is set"
            )
            result = False

        if (
            "bridge-create" in iface["host"]
            and iface["host"]["bridge-create"]
            and not "bridge" in iface["host"]
        ):
            msgs.append(f"tap {ifname} bridge-create can only be set if bridge is set")
            result = False

        if "mac" in iface["host"] and mac.is_multicast(iface["host"]["mac"]):
            msgs.append(
                f"tap {ifname} host MAC address {iface['host']['mac']} cannot be multicast"
            )
            result = False

    return result, msgs



# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""A vppcfg configuration module that exposes its semantic/syntax validators"""
from __future__ import (
    absolute_import,
    division,
    print_function,
)

import logging
import ipaddress
import os.path
import sys

try:
    import yamale
except ImportError:
    print("ERROR: install yamale manually: sudo pip install yamale")
    sys.exit(-2)
from yamale import validators

from .loopback import validate_loopbacks
from .bondethernet import validate_bondethernets
from .interface import validate_interfaces
from .bridgedomain import validate_bridgedomains
from .vxlan_tunnel import validate_vxlan_tunnels
from .tap import validate_taps
from .prefixlist import validate_prefixlists
from .acl import validate_acls
from .sflow import validate_sflow


class IPInterfaceWithPrefixLength(validators.Validator):
    """Custom IPAddress config - takes IP/prefixlen as input:
    192.0.2.1/29 or 2001:db8::1/64 are correct. The PrefixLength
    is required, and must be a number (0-32 for IPv4 and 0-128 for
    IPv6).
    """

    tag = "ip_interface"

    def _is_valid(self, value):
        try:
            _network = ipaddress.ip_interface(value)
        except ValueError:
            return False
        if not isinstance(value, str):
            return False
        if not "/" in value:
            return False
        elems = value.split("/")
        if not len(elems) == 2:
            return False
        if not elems[1].isnumeric():
            return False
        return True


class Validator:
    """The Validator class takes a schema filename (which may be None, in which
    case a built-in default is used), and a given YAML file represented as a string,
    and holds it against syntax and semantic validators, returning a tuple of (bool,list)
    where the boolean signals success/failure, and the list of strings are messages
    that were added when validating the YAML config.

    The purpose is to  ensure that the YAML file is both syntactically correct,
    which is ensured by Yamale, and semantically correct, which is ensured by a set
    of built-in validators, and user-added validators (see the add_validator() method).
    """

    def __init__(self, schema):
        self.logger = logging.getLogger("vppcfg.config")
        self.logger.addHandler(logging.NullHandler())

        self.schema = schema
        self.validators = [
            validate_bondethernets,
            validate_interfaces,
            validate_loopbacks,
            validate_bridgedomains,
            validate_vxlan_tunnels,
            validate_taps,
            validate_prefixlists,
            validate_acls,
            validate_sflow,
        ]

    def validate(self, yaml):
        """Validate the semantics of all YAML maps, by calling self.validators in turn,
        and then optionally calling validators that were added with add_validator()"""
        ret_retval = True
        ret_msgs = []
        if not yaml:
            return ret_retval, ret_msgs

        _validators = validators.DefaultValidators.copy()
        _validators[IPInterfaceWithPrefixLength.tag] = IPInterfaceWithPrefixLength
        if self.schema:
            fname = self.schema
            self.logger.debug(f"Validating against --schema {fname}")
        else:
            ## See setup.py data files that includes schema.yaml into the bundle
            self.logger.debug("Validating against built-in schema")
            fname = os.path.abspath(
                os.path.join(os.path.dirname(__file__), "..", "schema.yaml")
            )

        if not os.path.isfile(fname):
            self.logger.error(f"Cannot file schema file: {fname}")
            return False, ret_msgs

        try:
            schema = yamale.make_schema(fname, validators=_validators)
            data = yamale.make_data(content=str(yaml))
            yamale.validate(schema, data)
            self.logger.debug("Schema correctly validated by yamale")
        except yamale.YamaleError as err:
            ret_retval = False
            for result in err.results:
                for error in result.errors:
                    ret_msgs.extend([f"yamale: {error}"])
            return ret_retval, ret_msgs

        self.logger.debug("Validating Semantics...")

        for validator in self.validators:
            retval, msgs = validator(yaml)
            if msgs:
                ret_msgs.extend(msgs)
            if not retval:
                ret_retval = False

        if ret_retval:
            self.logger.debug("Semantics correctly validated")
        return ret_retval, ret_msgs

    def valid_config(self, yaml):
        """Validate the given YAML configuration in 'yaml' against syntax
        validation given in the yamale 'schema', and all semantic configs.

        Returns True if the configuration is valid, False otherwise.
        """

        retval, msgs = self.validate(yaml)
        if not retval:
            for msg in msgs:
                self.logger.error(msg)
            return False

        self.logger.info("Configuration validated successfully")
        return True

    def add_validator(self, func):
        """Add a validator function, which strictly takes the prototype
           rv, msgs = func(yaml)
        returning a Boolean success value in rv and a List of strings
        in msgs. The function will be passed the configuration YAML and
        gets to opine if it's valid or not.

        Note: will only be called iff Yamale syntax-check succeeded,
              and it will be called after all built-in validators.
        """
        self.validators.append(func)



"""module to help locate unittest resources"""

# Copyright (c) 2022 Ray Kinsella
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-

import os


class UnitTestYaml:
    """Helper classes to find and load unit test yaml resources"""

    def __init__(self, fpath):
        self.file = None
        self.filename = fpath
        self.resdir = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "../unittest")
        )

    def __enter__(self):
        self.file = open(
            os.path.join(self.resdir, self.filename), "r", encoding="utf-8"
        )
        return self.file

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.close()

    def __str__(self):
        return self.file



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A vppcfg configuration module that validates Linux Control Plane (lcp) elements"""


def get_lcps(yaml, interfaces=True, loopbacks=True, bridgedomains=True):
    """Returns a list of LCPs configured in the system. Optionally (de)select the different
    types of LCP. Return an empty list if there are none of the given type(s)."""

    ret = []
    if interfaces and "interfaces" in yaml:
        for _ifname, iface in yaml["interfaces"].items():
            if "lcp" in iface:
                ret.append(iface["lcp"])
            if "sub-interfaces" in iface:
                for _subid, sub_iface in iface["sub-interfaces"].items():
                    if "lcp" in sub_iface:
                        ret.append(sub_iface["lcp"])

    if loopbacks and "loopbacks" in yaml:
        for _ifname, iface in yaml["loopbacks"].items():
            if "lcp" in iface:
                ret.append(iface["lcp"])
    if bridgedomains and "bridgedomains" in yaml:
        for _ifname, iface in yaml["bridgedomains"].items():
            if "lcp" in iface:
                ret.append(iface["lcp"])

    return ret


def is_unique(yaml, lcpname):
    """Returns True if there is at most one occurence of the LCP name in the entire config."""

    lcps = get_lcps(yaml)
    return lcps.count(lcpname) < 2



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""
The functions in this file interact with the VPP API to modify certain
interface metadata.
"""

import logging
from .vppapi import VPPApi


class Applier(VPPApi):
    """The methods in the Applier class modify the running state in the VPP dataplane
    and will ensure that the local cache is consistent after creations and
    modifications."""

    # pylint: disable=unnecessary-pass

    def __init__(
        self,
        cfg,
        planner_cli,
        vpp_api_socket="/run/vpp/api.sock",
        vpp_json_dir=None,
        clientname="vppcfg",
    ):
        super().__init__(vpp_api_socket, vpp_json_dir, clientname)
        self.logger = logging.getLogger("vppcfg.applier")
        self.logger.addHandler(logging.NullHandler())
        self.cli = planner_cli

    def apply(self):
        """Apply the commands from self.cli to the cli_inband API call. Will eventually be
        replaced with actual API calls."""

        cli_calls = 0
        cli_success = 0
        for phase, cmds in self.cli.items():
            for cmd in cmds:
                cli_calls += 1

                self.logger.debug(f"{phase}: {cmd}")
                ret = self.cli_inband(cmd=cmd)
                self.logger.debug(f"Retval: {ret}")

                if ret is False:
                    self.logger.error("VPP returned error")
                elif ret.retval == 0:
                    cli_success += 1
                else:
                    self.logger.warning(f"VPP {cmd} returned {ret}")

        if cli_calls == 0:
            self.logger.info("Nothing to do")
            return True

        self.logger.info(f"VPP API calls: {cli_calls}, success: {cli_success}")
        if cli_calls != cli_success:
            self.logger.warning("Not all VPP calls were successful!")
            return False
        return True



#!/usr/bin/env python3
#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""This is a unit test suite for vppcfg"""
# pylint: disable=duplicate-code
import os
import sys
import glob
import re
import unittest
import yaml

try:
    from vppcfg.config import Validator
except ModuleNotFoundError:
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
    from vppcfg.config import Validator

try:
    import argparse
except ImportError:
    print("ERROR: install argparse manually: sudo pip install argparse")
    sys.exit(-2)


def example_validator(_yaml):
    """A simple example validator that takes the YAML configuration file as an input,
    and returns a tuple of rv (return value, True is success), and a list of string
    messages to the validation framework."""
    return True, []


class YAMLTest(unittest.TestCase):
    """This test suite takes a YAML configuration file and holds it against the syntax
    (Yamale) and semantic validators, returning errors in case of validation failures.
    """

    def __init__(self, testName, yaml_filename, yaml_schema):
        # calling the super class init varies for different python versions.  This works for 2.7
        super().__init__(testName)
        self.yaml_filename = yaml_filename
        self.yaml_schema = yaml_schema

    def test_yaml(self):
        """The test executor"""
        test = None
        cfg = None
        ncount = 0
        with open(self.yaml_filename, "r", encoding="utf-8") as file:
            for data in yaml.load_all(file, Loader=yaml.Loader):
                if ncount == 0:
                    test = data
                    ncount += 1
                elif ncount == 1:
                    cfg = data
                    ncount += 1
        self.assertEqual(ncount, 2)
        self.assertIsNotNone(test)
        if not cfg:
            return

        validator = Validator(schema=self.yaml_schema)
        _rv, msgs = validator.validate(cfg)

        msgs_expected = []
        if (
            "test" in test
            and "errors" in test["test"]
            and "expected" in test["test"]["errors"]
        ):
            msgs_expected = test["test"]["errors"]["expected"]

        fail = False
        for msg in msgs:
            this_msg_expected = False
            for expected in msgs_expected:
                if re.match(expected, msg):
                    this_msg_expected = True
                    break
            if not this_msg_expected:
                print(
                    f"{self.yaml_filename}: Unexpected message: {msg}", file=sys.stderr
                )
                fail = True

        count = 0
        if (
            "test" in test
            and "errors" in test["test"]
            and "count" in test["test"]["errors"]
        ):
            count = test["test"]["errors"]["count"]

        if len(msgs) != count:
            print(
                f"{self.yaml_filename}: Unexpected error count {len(msgs)} (expecting {int(count)})",
                file=sys.stderr,
            )
        self.assertEqual(len(msgs), count)
        self.assertFalse(fail)

        return


if __name__ == "__main__":
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        "-t",
        "--test",
        dest="test",
        type=str,
        nargs="+",
        default=["unittest/yaml/*.yaml"],
        help="""YAML test file(s)""",
    )
    parser.add_argument(
        "-s",
        "--schema",
        dest="schema",
        type=str,
        default="./schema.yaml",
        help="""YAML schema validation file""",
    )
    parser.add_argument(
        "-d",
        "--debug",
        dest="debug",
        action="store_true",
        help="""Enable debug, default False""",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        help="""Be quiet (only log warnings/errors), default False""",
    )

    args = parser.parse_args()
    if args.debug:
        VERBOSITY = 2
    elif args.quiet:
        VERBOSITY = 0
    else:
        VERBOSITY = 1
    yaml_suite = unittest.TestSuite()
    for pattern in args.test:
        for fn in glob.glob(pattern):
            yaml_suite.addTest(
                YAMLTest("test_yaml", yaml_filename=fn, yaml_schema=args.schema)
            )
    yaml_ok = (
        unittest.TextTestRunner(verbosity=VERBOSITY, buffer=True)
        .run(yaml_suite)
        .wasSuccessful()
    )

    tests = unittest.TestLoader().discover(start_dir=".", pattern="test_*.py")
    unit_ok = (
        unittest.TextTestRunner(verbosity=VERBOSITY, buffer=True)
        .run(tests)
        .wasSuccessful()
    )

    retval = 0
    if not yaml_ok:
        retval -= 1
    if not unit_ok:
        retval -= 2
    sys.exit(retval)



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""Unit tests for interfaces"""
import unittest
import yaml
from . import interface
from .unittestyaml import UnitTestYaml


class TestInterfaceMethods(unittest.TestCase):
    def setUp(self):
        with UnitTestYaml("test_interface.yaml") as f:
            self.cfg = yaml.load(f, Loader=yaml.FullLoader)

    def test_enumerators(self):
        ifs = interface.get_interfaces(self.cfg)
        self.assertEqual(len(ifs), 27)
        self.assertIn("GigabitEthernet1/0/1", ifs)
        self.assertIn("GigabitEthernet1/0/1.200", ifs)

        ifs = interface.get_sub_interfaces(self.cfg)
        self.assertEqual(len(ifs), 17)
        self.assertNotIn("GigabitEthernet1/0/1", ifs)
        self.assertIn("GigabitEthernet1/0/1.200", ifs)
        self.assertIn("GigabitEthernet1/0/1.201", ifs)
        self.assertIn("GigabitEthernet1/0/1.202", ifs)
        self.assertIn("GigabitEthernet1/0/1.203", ifs)

        ifs = interface.get_qinx_interfaces(self.cfg)
        self.assertEqual(len(ifs), 3)
        self.assertNotIn("GigabitEthernet1/0/1.200", ifs)
        self.assertNotIn("GigabitEthernet1/0/1.202", ifs)
        self.assertIn("GigabitEthernet1/0/1.201", ifs)
        self.assertIn("GigabitEthernet1/0/1.203", ifs)

        ifs = interface.get_l2xc_interfaces(self.cfg)
        self.assertEqual(len(ifs), 3)
        self.assertIn("GigabitEthernet3/0/0", ifs)
        self.assertIn("GigabitEthernet3/0/1", ifs)
        self.assertIn("GigabitEthernet3/0/2.100", ifs)
        self.assertNotIn("GigabitEthernet3/0/2.200", ifs)

        target_ifs = interface.get_l2xc_target_interfaces(self.cfg)
        self.assertEqual(len(target_ifs), 3)
        self.assertIn("GigabitEthernet3/0/0", target_ifs)
        self.assertIn("GigabitEthernet3/0/1", target_ifs)
        self.assertNotIn("GigabitEthernet3/0/2.100", target_ifs)
        self.assertIn("GigabitEthernet3/0/2.200", target_ifs)

        ## Since l2xc cannot connect to itself, and the target must exist,
        ## it follows that the same number of l2xc target interfaces must exist.
        self.assertEqual(len(target_ifs), len(ifs))

    def test_mtu(self):
        self.assertEqual(interface.get_mtu(self.cfg, "GigabitEthernet1/0/1"), 9216)
        self.assertEqual(interface.get_mtu(self.cfg, "GigabitEthernet1/0/1.200"), 9000)
        self.assertEqual(interface.get_mtu(self.cfg, "GigabitEthernet1/0/1.201"), 9216)

    def test_encapsulation(self):
        self.assertTrue(
            interface.valid_encapsulation(self.cfg, "GigabitEthernet1/0/1.200")
        )
        self.assertTrue(
            interface.unique_encapsulation(self.cfg, "GigabitEthernet1/0/1.200")
        )
        self.assertEqual(
            interface.get_encapsulation(self.cfg, "GigabitEthernet1/0/1.200"),
            {"dot1q": 1000, "dot1ad": 0, "inner-dot1q": 0, "exact-match": False},
        )
        self.assertEqual(
            interface.get_encapsulation(self.cfg, "GigabitEthernet1/0/1.201"),
            {"dot1q": 1000, "dot1ad": 0, "inner-dot1q": 1234, "exact-match": False},
        )
        self.assertEqual(
            interface.get_encapsulation(self.cfg, "GigabitEthernet1/0/1.202"),
            {"dot1q": 0, "dot1ad": 1000, "inner-dot1q": 0, "exact-match": False},
        )
        self.assertEqual(
            interface.get_encapsulation(self.cfg, "GigabitEthernet1/0/1.203"),
            {"dot1q": 0, "dot1ad": 1000, "inner-dot1q": 1000, "exact-match": True},
        )

        self.assertFalse(
            interface.valid_encapsulation(self.cfg, "GigabitEthernet1/0/0.100")
        )
        self.assertFalse(
            interface.valid_encapsulation(self.cfg, "GigabitEthernet1/0/0.101")
        )

        self.assertFalse(
            interface.unique_encapsulation(self.cfg, "GigabitEthernet1/0/0.102")
        )
        self.assertFalse(
            interface.unique_encapsulation(self.cfg, "GigabitEthernet1/0/0.103")
        )

    def test_has_sub(self):
        self.assertTrue(interface.has_sub(self.cfg, "GigabitEthernet1/0/1"))
        self.assertFalse(interface.has_sub(self.cfg, "GigabitEthernet1/0/1.200"))
        self.assertFalse(interface.has_sub(self.cfg, "GigabitEthernet2/0/0"))
        self.assertFalse(interface.has_sub(self.cfg, "GigabitEthernet3/0/0"))

    def test_is_sub(self):
        self.assertFalse(interface.is_sub(self.cfg, "GigabitEthernet1/0/1"))
        self.assertTrue(interface.is_sub(self.cfg, "GigabitEthernet1/0/1.200"))

    def test_is_qinx(self):
        self.assertFalse(interface.is_qinx(self.cfg, "GigabitEthernet1/0/1"))
        self.assertFalse(interface.is_qinx(self.cfg, "GigabitEthernet1/0/1.200"))
        self.assertFalse(interface.is_qinx(self.cfg, "GigabitEthernet1/0/1.202"))

        self.assertTrue(interface.is_qinx(self.cfg, "GigabitEthernet1/0/1.201"))
        self.assertTrue(interface.is_qinx(self.cfg, "GigabitEthernet1/0/1.203"))

    def test_has_lcp(self):
        self.assertTrue(interface.has_lcp(self.cfg, "GigabitEthernet1/0/1"))
        self.assertFalse(interface.has_lcp(self.cfg, "GigabitEthernet1/0/0"))

    def test_get_lcp(self):
        self.assertIsNone(interface.get_lcp(self.cfg, "GigabitEthernet1/0/0"))
        self.assertIsNone(interface.get_lcp(self.cfg, "GigabitEthernet1/0/0.100"))

        self.assertEqual(interface.get_lcp(self.cfg, "GigabitEthernet1/0/1"), "e1")
        self.assertEqual(interface.get_lcp(self.cfg, "GigabitEthernet1/0/1.100"), "foo")
        self.assertEqual(
            interface.get_lcp(self.cfg, "GigabitEthernet1/0/1.101"), "e1.100"
        )
        self.assertEqual(
            interface.get_lcp(self.cfg, "GigabitEthernet1/0/1.102"), "e1.100.100"
        )

        self.assertIsNone(interface.get_lcp(self.cfg, "GigabitEthernet1/0/1.200"))
        self.assertIsNone(interface.get_lcp(self.cfg, "GigabitEthernet1/0/1.201"))
        self.assertIsNone(interface.get_lcp(self.cfg, "GigabitEthernet1/0/1.202"))
        self.assertIsNone(interface.get_lcp(self.cfg, "GigabitEthernet1/0/1.203"))

    def test_address(self):
        self.assertFalse(interface.has_address(self.cfg, "GigabitEthernet1/0/0"))
        self.assertFalse(interface.has_address(self.cfg, "GigabitEthernet1/0/0.100"))

        self.assertTrue(interface.has_address(self.cfg, "GigabitEthernet1/0/1"))
        self.assertTrue(interface.has_address(self.cfg, "GigabitEthernet1/0/1.100"))

    def test_lx2c(self):
        l2xc_ifs = interface.get_l2xc_interfaces(self.cfg)
        l2xc_target_ifs = interface.get_l2xc_target_interfaces(self.cfg)

        self.assertIn("GigabitEthernet3/0/0", l2xc_ifs)
        self.assertIn("GigabitEthernet3/0/0", l2xc_target_ifs)
        self.assertTrue(interface.is_l2xc_interface(self.cfg, "GigabitEthernet3/0/0"))
        self.assertTrue(
            interface.is_l2xc_target_interface(self.cfg, "GigabitEthernet3/0/0")
        )

        self.assertNotIn("GigabitEthernet2/0/0", l2xc_ifs)
        self.assertNotIn("GigabitEthernet2/0/0", l2xc_target_ifs)
        self.assertFalse(interface.is_l2xc_interface(self.cfg, "GigabitEthernet2/0/0"))
        self.assertFalse(
            interface.is_l2xc_target_interface(self.cfg, "GigabitEthernet2/0/0")
        )

    def test_l2(self):
        self.assertTrue(interface.is_l2(self.cfg, "GigabitEthernet3/0/0"))
        self.assertFalse(interface.is_l2(self.cfg, "GigabitEthernet1/0/0"))
        self.assertTrue(interface.is_l2(self.cfg, "GigabitEthernet3/0/2.100"))
        self.assertTrue(interface.is_l2(self.cfg, "GigabitEthernet3/0/2.200"))

    def test_l3(self):
        self.assertTrue(interface.is_l3(self.cfg, "GigabitEthernet1/0/0"))
        self.assertFalse(interface.is_l3(self.cfg, "GigabitEthernet3/0/0"))

    def test_get_by_lcp_name(self):
        ifname, iface = interface.get_by_lcp_name(self.cfg, "notexist")
        self.assertIsNone(ifname)
        self.assertIsNone(iface)

        ifname, iface = interface.get_by_lcp_name(self.cfg, "e1.100.100")
        self.assertEqual(ifname, "GigabitEthernet1/0/1.102")
        ifname, iface = interface.get_by_lcp_name(self.cfg, "e2")
        self.assertEqual(ifname, "GigabitEthernet2/0/0")

    def test_get_by_name(self):
        ifname, iface = interface.get_by_name(self.cfg, "GigabitEthernet1/0/1.201")
        self.assertEqual(ifname, "GigabitEthernet1/0/1.201")
        self.assertIsNotNone(iface)
        encap = interface.get_encapsulation(self.cfg, ifname)
        self.assertEqual(
            encap,
            {"dot1q": 1000, "dot1ad": 0, "inner-dot1q": 1234, "exact-match": False},
        )

        ifname, iface = interface.get_by_name(self.cfg, "GigabitEthernet1/0/1.1")
        self.assertIsNone(ifname)
        self.assertIsNone(iface)

    def test_get_parent_by_name(self):
        ifname, iface = interface.get_parent_by_name(
            self.cfg, "GigabitEthernet1/0/1.201"
        )
        self.assertEqual(ifname, "GigabitEthernet1/0/1")
        self.assertIsNotNone(iface)
        self.assertNotIn("encapsulation", iface)

        ifname, iface = interface.get_parent_by_name(
            self.cfg, "GigabitEthernet1/0/1.200"
        )
        self.assertEqual(ifname, "GigabitEthernet1/0/1")
        self.assertIsNotNone(iface)
        self.assertNotIn("encapsulation", iface)

        ifname, iface = interface.get_parent_by_name(self.cfg, "GigabitEthernet1/0/1")
        self.assertIsNone(ifname)
        self.assertIsNone(iface)

        ifname, iface = interface.get_parent_by_name(self.cfg, None)
        self.assertIsNone(ifname)
        self.assertIsNone(iface)

    def test_get_qinx_parent_by_name(self):
        self.assertIsNotNone(
            interface.get_qinx_parent_by_name(self.cfg, "GigabitEthernet1/0/1.202")
        )
        self.assertIsNotNone(
            interface.get_qinx_parent_by_name(self.cfg, "GigabitEthernet1/0/1.203")
        )

        ifname, iface = interface.get_qinx_parent_by_name(
            self.cfg, "GigabitEthernet1/0/1"
        )
        self.assertIsNone(iface)
        self.assertIsNone(ifname)

        ifname, iface = interface.get_qinx_parent_by_name(
            self.cfg, "GigabitEthernet1/0/1.100"
        )
        self.assertIsNone(iface)
        self.assertIsNone(ifname)

        ifname, iface = interface.get_qinx_parent_by_name(
            self.cfg, "GigabitEthernet1/0/1.200"
        )
        self.assertIsNone(iface)
        self.assertIsNone(ifname)

        ifname, iface = interface.get_qinx_parent_by_name(
            self.cfg, "GigabitEthernet1/0/1.201"
        )
        self.assertEqual(ifname, "GigabitEthernet1/0/1.200")

    def test_get_phys(self):
        phys = interface.get_phys(self.cfg)
        self.assertEqual(len(phys), 10)
        self.assertIn("GigabitEthernet1/0/0", phys)
        self.assertNotIn("GigabitEthernet1/0/0.100", phys)

    def test_is_phy(self):
        self.assertTrue(interface.is_phy(self.cfg, "GigabitEthernet1/0/0"))
        self.assertFalse(interface.is_phy(self.cfg, "GigabitEthernet1/0/0.100"))

    def test_get_admin_state(self):
        self.assertFalse(interface.get_admin_state(self.cfg, "notexist"))
        self.assertFalse(interface.get_admin_state(self.cfg, "GigabitEthernet2/0/0"))
        self.assertTrue(interface.get_admin_state(self.cfg, "GigabitEthernet1/0/0"))
        self.assertTrue(interface.get_admin_state(self.cfg, "GigabitEthernet1/0/0.101"))
        self.assertFalse(
            interface.get_admin_state(self.cfg, "GigabitEthernet1/0/0.102")
        )

    def test_is_mpls(self):
        self.assertTrue(interface.is_mpls(self.cfg, "GigabitEthernet1/0/1"))
        self.assertTrue(interface.is_mpls(self.cfg, "GigabitEthernet1/0/1.101"))
        self.assertFalse(interface.is_mpls(self.cfg, "GigabitEthernet1/0/0"))
        self.assertFalse(interface.is_mpls(self.cfg, "GigabitEthernet1/0/0.100"))
        self.assertFalse(interface.is_mpls(self.cfg, "notexist"))

    def test_get_unnumbered_interfaces(self):
        unnumbered_ifaces = interface.get_unnumbered_interfaces(self.cfg)
        self.assertEqual(len(unnumbered_ifaces), 5)
        self.assertNotIn("GigabitEthernet4/0/0", unnumbered_ifaces)
        self.assertIn("GigabitEthernet4/0/1", unnumbered_ifaces)
        self.assertNotIn("GigabitEthernet4/0/3.100", unnumbered_ifaces)
        self.assertIn("GigabitEthernet4/0/3.101", unnumbered_ifaces)

    def test_is_unnumbered(self):
        self.assertFalse(interface.is_unnumbered(self.cfg, "GigabitEthernet4/0/0"))
        self.assertTrue(interface.is_unnumbered(self.cfg, "GigabitEthernet4/0/1"))
        self.assertTrue(interface.is_unnumbered(self.cfg, "GigabitEthernet4/0/2"))
        self.assertFalse(interface.is_unnumbered(self.cfg, "GigabitEthernet4/0/3.100"))
        self.assertTrue(interface.is_unnumbered(self.cfg, "GigabitEthernet4/0/3.101"))
        self.assertTrue(interface.is_unnumbered(self.cfg, "GigabitEthernet4/0/3.102"))
        self.assertTrue(interface.is_unnumbered(self.cfg, "GigabitEthernet4/0/3.103"))



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A vppcfg configuration module that validates vxlan_tunnels"""
import logging
import ipaddress


def get_by_name(yaml, ifname):
    """Return the VXLAN by name, if it exists. Return None otherwise."""
    try:
        if ifname in yaml["vxlan_tunnels"]:
            return ifname, yaml["vxlan_tunnels"][ifname]
    except KeyError:
        pass
    return None, None


def is_vxlan_tunnel(yaml, ifname):
    """Returns True if the interface name is an existing VXLAN Tunnel."""
    ifname, iface = get_by_name(yaml, ifname)
    return iface is not None


def vni_unique(yaml, vni):
    """Return True if the VNI is unique amongst all VXLANs"""
    if not "vxlan_tunnels" in yaml:
        return True

    ncount = 0
    for _ifname, iface in yaml["vxlan_tunnels"].items():
        if iface["vni"] == vni:
            ncount = ncount + 1

    return ncount < 2


def get_vxlan_tunnels(yaml):
    """Returns a list of all VXLAN tunnel interface names."""
    ret = []
    if not "vxlan_tunnels" in yaml:
        return ret

    for ifname, _iface in yaml["vxlan_tunnels"].items():
        ret.append(ifname)
    return ret


def validate_vxlan_tunnels(yaml):
    """Validate the semantics of all YAML 'vxlan_tunnels' entries"""
    result = True
    msgs = []
    logger = logging.getLogger("vppcfg.config")
    logger.addHandler(logging.NullHandler())

    if not "vxlan_tunnels" in yaml:
        return result, msgs

    for ifname, iface in yaml["vxlan_tunnels"].items():
        logger.debug(f"vxlan_tunnel {ifname}: {iface}")
        instance = int(ifname[12:])
        if instance > 2147483647:
            msgs.append(
                f"vxlan_tunnel {ifname} has instance {int(instance)} which is too large"
            )
            result = False

        vni = iface["vni"]
        if not vni_unique(yaml, vni):
            msgs.append(f"vxlan_tunnel {ifname} VNI {int(vni)} is not unique")
            result = False
        local = ipaddress.ip_address(iface["local"])
        remote = ipaddress.ip_address(iface["remote"])
        if local.version != remote.version:
            msgs.append(
                f"vxlan_tunnel {ifname} local and remote are not the same address family"
            )
            result = False

    return result, msgs



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""
The functions in this file interact with the VPP API to retrieve certain
interface metadata and write it to a YAML file.
"""

import sys
import yaml
from vppcfg.config import bondethernet
from .vppapi import VPPApi


class Dumper(VPPApi):
    """The Dumper class first reads the configuration from a running VPP Dataplane
    by using a set of (readonly) API getters, and then emits the configuration as
    a YAML file with its write() method, either to stdout or to a filename.

    Note that not all running VPP configs are "valid" in vppcfg's eyes. It is not
    guaranteed that the output of the Dumper() will stand validation."""

    def __init__(
        self,
        vpp_api_socket="/run/vpp/api.sock",
        vpp_json_dir=None,
        clientname="vppcfg",
    ):
        VPPApi.__init__(self, vpp_api_socket, vpp_json_dir, clientname)

    def write(self, outfile):
        """Emit the configuration to either stdout (outfile=='-') or a filename"""
        config = self.cache_to_config()
        if outfile and outfile == "-":
            file = sys.stdout
            outfile = "(stdout)"
            print(yaml.dump(config), file=file)
        else:
            with open(outfile, "w", encoding="utf-8") as file:
                print(yaml.dump(config), file=file)
        self.logger.info(f"Wrote YAML config to {outfile}")

    def cache_to_config(self):
        """Convert the VPP configuration cache (previously read by readconfig() into
        a YAML representation."""
        config = {
            "loopbacks": {},
            "bondethernets": {},
            "interfaces": {},
            "bridgedomains": {},
            "vxlan_tunnels": {},
            "taps": {},
            "prefixlists": {},
            "acls": {},
            "sflow": {},
        }
        for idx, bond_iface in self.cache["bondethernets"].items():
            bond = {"description": ""}
            if bond_iface.sw_if_index in self.cache["bondethernet_members"]:
                members = [
                    self.cache["interfaces"][x].interface_name
                    for x in self.cache["bondethernet_members"][bond_iface.sw_if_index]
                ]
                if len(members) > 0:
                    bond["interfaces"] = members

            mode = bondethernet.int_to_mode(bond_iface.mode)
            bond["mode"] = mode
            if mode in ["xor", "lacp"]:
                bond["load-balance"] = bondethernet.int_to_lb(bond_iface.lb)
            iface = self.cache["interfaces"][bond_iface.sw_if_index]
            bond["mac"] = str(iface.l2_address)
            config["bondethernets"][iface.interface_name] = bond

        for numtags in [0, 1, 2]:
            for idx, iface in self.cache["interfaces"].items():
                if iface.sub_number_of_tags != numtags:
                    continue

                if iface.interface_dev_type == "Loopback":
                    if iface.sub_id > 0:
                        self.logger.warning(
                            f"Refusing to export sub-interfaces of loopback devices ({iface.interface_name})"
                        )
                        continue
                    loop = {"description": ""}
                    loop["mtu"] = iface.mtu[0]
                    loop["mac"] = str(iface.l2_address)
                    if iface.sw_if_index in self.cache["lcps"]:
                        loop["lcp"] = self.cache["lcps"][iface.sw_if_index].host_if_name
                    if iface.sw_if_index in self.cache["interface_unnumbered"]:
                        target = self.cache["interface_unnumbered"][iface.sw_if_index]
                        loop["unnumbered"] = self.cache["interfaces"][
                            target
                        ].interface_name
                    if iface.sw_if_index in self.cache["interface_addresses"]:
                        if (
                            len(self.cache["interface_addresses"][iface.sw_if_index])
                            > 0
                        ):
                            loop["addresses"] = self.cache["interface_addresses"][
                                iface.sw_if_index
                            ]
                    if iface.sw_if_index in self.cache["interface_mpls"]:
                        loop["mpls"] = self.cache["interface_mpls"][iface.sw_if_index]
                    config["loopbacks"][iface.interface_name] = loop
                elif iface.interface_dev_type in [
                    "bond",
                    "VXLAN",
                    "dpdk",
                    "af-packet",
                    "virtio",
                    "pg",
                ]:
                    i = {"description": ""}
                    if iface.sw_if_index in self.cache["lcps"]:
                        i["lcp"] = self.cache["lcps"][iface.sw_if_index].host_if_name
                    if iface.sw_if_index in self.cache["interface_unnumbered"]:
                        target = self.cache["interface_unnumbered"][iface.sw_if_index]
                        i["unnumbered"] = self.cache["interfaces"][
                            target
                        ].interface_name
                    if iface.sw_if_index in self.cache["interface_addresses"]:
                        if (
                            len(self.cache["interface_addresses"][iface.sw_if_index])
                            > 0
                        ):
                            i["addresses"] = self.cache["interface_addresses"][
                                iface.sw_if_index
                            ]
                    if iface.sw_if_index in self.cache["interface_mpls"]:
                        i["mpls"] = self.cache["interface_mpls"][iface.sw_if_index]
                    if iface.sw_if_index in self.cache["l2xcs"]:
                        l2xc = self.cache["l2xcs"][iface.sw_if_index]
                        i["l2xc"] = self.cache["interfaces"][
                            l2xc.tx_sw_if_index
                        ].interface_name
                    if (
                        not self.cache["interfaces"][idx].flags & 1
                    ):  # IF_STATUS_API_FLAG_ADMIN_UP
                        i["state"] = "down"

                    if (
                        iface.interface_dev_type in ["dpdk", "af-packet"]
                        and iface.sub_number_of_tags == 0
                    ):
                        i["mac"] = str(iface.l2_address)

                    if self.tap_is_lcp(iface.interface_name):
                        continue

                    i["mtu"] = iface.mtu[0]
                    if iface.sub_number_of_tags == 0:
                        config["interfaces"][iface.interface_name] = i
                        continue

                    encap = {}
                    if iface.sub_if_flags & 8:
                        encap["dot1ad"] = iface.sub_outer_vlan_id
                    else:
                        encap["dot1q"] = iface.sub_outer_vlan_id
                    if iface.sub_inner_vlan_id > 0:
                        encap["inner-dot1q"] = iface.sub_inner_vlan_id
                    encap["exact-match"] = bool(iface.sub_if_flags & 16)
                    i["encapsulation"] = encap

                    sup_iface = self.cache["interfaces"][iface.sup_sw_if_index]
                    if iface.mtu[0] > 0:
                        i["mtu"] = iface.mtu[0]
                    else:
                        i["mtu"] = sup_iface.mtu[0]
                    if (
                        not "sub-interfaces"
                        in config["interfaces"][sup_iface.interface_name]
                    ):
                        config["interfaces"][sup_iface.interface_name][
                            "sub-interfaces"
                        ] = {}
                    config["interfaces"][sup_iface.interface_name]["sub-interfaces"][
                        iface.sub_id
                    ] = i
        if iface.interface_dev_type in ["dpdk", "af-packet"]:
            config["interfaces"][iface.interface_name][
                "device-type"
            ] = iface.interface_dev_type

        for idx, iface in self.cache["vxlan_tunnels"].items():
            vpp_iface = self.cache["interfaces"][iface.sw_if_index]
            vxlan = {
                "description": "",
                "vni": int(iface.vni),
                "local": str(iface.src_address),
                "remote": str(iface.dst_address),
            }
            config["vxlan_tunnels"][vpp_iface.interface_name] = vxlan

        for idx, iface in self.cache["taps"].items():
            vpp_tap = self.cache["taps"][iface.sw_if_index]
            vpp_iface = self.cache["interfaces"][vpp_tap.sw_if_index]
            if self.tap_is_lcp(vpp_iface.interface_name):
                continue

            tap = {
                "description": "",
                "tx-ring-size": vpp_tap.tx_ring_sz,
                "rx-ring-size": vpp_tap.rx_ring_sz,
                "host": {
                    "mac": str(vpp_tap.host_mac_addr),
                    "name": vpp_tap.host_if_name,
                },
            }
            if vpp_tap.host_mtu_size > 0:
                tap["host"]["mtu"] = vpp_tap.host_mtu_size
            if vpp_tap.host_namespace:
                tap["host"]["namespace"] = vpp_tap.host_namespace
            if vpp_tap.host_bridge:
                tap["host"]["bridge"] = vpp_tap.host_bridge
            config["taps"][vpp_iface.interface_name] = tap

        for idx, iface in self.cache["bridgedomains"].items():
            bridge_name = f"bd{int(idx)}"
            mtu = 1500
            bridge = {"description": ""}
            settings = {}
            settings["learn"] = iface.learn
            settings["unicast-flood"] = iface.flood
            settings["unknown-unicast-flood"] = iface.uu_flood
            settings["unicast-forward"] = iface.forward
            settings["arp-termination"] = iface.arp_term
            settings["arp-unicast-forward"] = iface.arp_ufwd
            settings["mac-age-minutes"] = int(iface.mac_age)
            bridge["settings"] = settings

            bvi = None
            if iface.bvi_sw_if_index != 2**32 - 1:
                bvi = self.cache["interfaces"][iface.bvi_sw_if_index]
                mtu = bvi.mtu[0]
                bridge["bvi"] = bvi.interface_name
            members = []
            for member in iface.sw_if_details:
                if (
                    bvi
                    and bvi.interface_name
                    == self.cache["interfaces"][member.sw_if_index].interface_name
                ):
                    continue
                members.append(
                    self.cache["interfaces"][member.sw_if_index].interface_name
                )
                mtu = self.cache["interfaces"][member.sw_if_index].mtu[0]
            if len(members) > 0:
                bridge["interfaces"] = members
            bridge["mtu"] = mtu
            config["bridgedomains"][bridge_name] = bridge
        for idx, acl in self.cache["acls"].items():
            aclname = f"vppacl{acl.acl_index}"

            descr = acl.tag.replace('"', "").replace("'", "")
            if descr != acl.tag:
                self.logger.warning(
                    f"acl tag {acl.tag} has invalid characters, stripping"
                )
            descr = "tag " + descr
            config_acl = {"description": descr, "terms": []}
            terms = 0
            for acl_rule in acl.r:
                terms += 1
                action = "deny"
                if acl_rule.is_permit == 1:
                    action = "permit"
                elif acl_rule.is_permit == 2:
                    action = "permit+reflect"

                config_term = {
                    "action": action,
                    "source": str(acl_rule.src_prefix),
                    "destination": str(acl_rule.dst_prefix),
                }
                if acl_rule.proto == 0:
                    pass
                elif acl_rule.proto in [1, 58]:
                    if acl_rule.proto == 1:
                        config_term["protocol"] = "icmp"
                    else:
                        config_term["protocol"] = "ipv6-icmp"
                    maxval = acl_rule.srcport_or_icmptype_last
                    if maxval > 255:
                        self.logger.warning(
                            f"icmp type > 255 on acl {acl.acl_index} term {terms}"
                        )
                        maxval = 255
                    if acl_rule.srcport_or_icmptype_first == maxval:
                        config_term["icmp-type"] = int(
                            acl_rule.srcport_or_icmptype_first
                        )
                    else:
                        config_term["icmp-type"] = (
                            f"{acl_rule.srcport_or_icmptype_first}-{maxval}"
                        )

                    maxval = acl_rule.dstport_or_icmpcode_last
                    if maxval > 255:
                        self.logger.warning(
                            f"icmp code > 255 on acl {acl.acl_index} term {terms}"
                        )
                        maxval = 255
                    if acl_rule.dstport_or_icmpcode_first == maxval:
                        config_term["icmp-code"] = int(
                            acl_rule.dstport_or_icmpcode_first
                        )
                    else:
                        config_term["icmp-code"] = (
                            f"{acl_rule.dstport_or_icmpcode_first}-{maxval}"
                        )
                elif acl_rule.proto in [6, 17]:
                    if acl_rule.proto == 6:
                        config_term["protocol"] = "tcp"
                    else:
                        config_term["protocol"] = "udp"
                    if (
                        acl_rule.srcport_or_icmptype_first
                        == acl_rule.srcport_or_icmptype_last
                    ):
                        config_term["source-port"] = int(
                            acl_rule.srcport_or_icmptype_first
                        )
                    else:
                        config_term["source-port"] = (
                            f"{acl_rule.srcport_or_icmptype_first}-{acl_rule.srcport_or_icmptype_last}"
                        )
                    if (
                        acl_rule.dstport_or_icmpcode_first
                        == acl_rule.dstport_or_icmpcode_last
                    ):
                        config_term["destination-port"] = int(
                            acl_rule.dstport_or_icmpcode_first
                        )
                    else:
                        config_term["destination-port"] = (
                            f"{acl_rule.dstport_or_icmpcode_first}-{acl_rule.dstport_or_icmpcode_last}"
                        )
                else:
                    config_term["protocol"] = int(acl_rule.proto)

                config_acl["terms"].append(config_term)

            config["acls"][aclname] = config_acl

        config["sflow"] = self.cache["sflow"]
        for hw_if_index in self.cache["interface_sflow"]:
            vpp_iface = self.cache["interfaces"][hw_if_index]
            config["interfaces"][vpp_iface.interface_name]["sflow"] = True

        return config



#
# Copyright (c) 2023 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A vppcfg configuration module that validates prefixlists"""
import logging
import ipaddress


def get_prefixlists(yaml):
    """Return a list of all prefixlists."""
    ret = []
    if "prefixlists" in yaml:
        for plname, _pl in yaml["prefixlists"].items():
            ret.append(plname)
    return ret


def get_by_name(yaml, plname):
    """Return the prefixlist by name, if it exists. Return None otherwise."""
    try:
        if plname in yaml["prefixlists"]:
            return plname, yaml["prefixlists"][plname]
    except KeyError:
        pass
    return None, None


def get_network_list(yaml, plname, want_ipv4=True, want_ipv6=True):
    """Returns a list of 0 or more ip_network elements, that represent the members
    in a prefixlist of given name. Return the empty list if the prefixlist doesn't
    exist. Optionally, want_ipv4 or want_ipv6 can be set to False to filter the list."""
    ret = []
    plname, plist = get_by_name(yaml, plname)
    if not plist:
        return ret
    for member in plist["members"]:
        ipn = ipaddress.ip_network(member, strict=False)
        if ipn.version == 4 and want_ipv4:
            ret.append(ipn)
        if ipn.version == 6 and want_ipv6:
            ret.append(ipn)
    return ret


def count(yaml, plname):
    """Return the number of IPv4 and IPv6 entries in the prefixlist.
    Returns 0, 0 if it doesn't exist"""
    ipv4, ipv6 = 0, 0

    plname, plist = get_by_name(yaml, plname)
    if not plist:
        return 0, 0
    for member in plist["members"]:
        ipn = ipaddress.ip_network(member, strict=False)
        if ipn.version == 4:
            ipv4 += 1
        elif ipn.version == 6:
            ipv6 += 1

    return ipv4, ipv6


def count_ipv4(yaml, plname):
    """Return the number of IPv4 entries in the prefixlist."""
    ipv4, _ = count(yaml, plname)
    return ipv4


def count_ipv6(yaml, plname):
    """Return the number of IPv6 entries in the prefixlist."""
    _, ipv6 = count(yaml, plname)
    return ipv6


def has_ipv4(yaml, plname):
    """Return True if the prefixlist has at least one IPv4 entry."""
    ipv4, _ = count(yaml, plname)
    return ipv4 > 0


def has_ipv6(yaml, plname):
    """Return True if the prefixlist has at least one IPv6 entry."""
    _, ipv6 = count(yaml, plname)
    return ipv6 > 0


def is_empty(yaml, plname):
    """Return True if the prefixlist has no entries."""
    ipv4, ipv6 = count(yaml, plname)
    return ipv4 + ipv6 == 0


def validate_prefixlists(yaml):
    """Validate the semantics of all YAML 'prefixlists' entries"""
    result = True
    msgs = []
    logger = logging.getLogger("vppcfg.config")
    logger.addHandler(logging.NullHandler())

    if not "prefixlists" in yaml:
        return result, msgs

    for plname, plist in yaml["prefixlists"].items():
        logger.debug(f"prefixlist {plname}: {plist}")
        if plname in ["any"]:
            ## Note: ACL 'source' and 'destination', when they are empty, will resolve
            ## to 'any', and can thus never refer to a prefixlist called 'any'.
            msgs.append(f"prefixlist {plname} is a reserved name")
            result = False

        members = 0
        for pl_member in plist["members"]:
            members += 1
            logger.debug(f"prefixlist {plname} member {members} is {pl_member}")

    return result, msgs



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""Unit tests for MAC addresses"""
import unittest
from . import mac


class TestMACMethods(unittest.TestCase):
    def test_is_valid(self):
        self.assertTrue(mac.is_valid("00:01:02:03:04:05"))
        self.assertTrue(mac.is_valid("00-01-02-03-04-05"))
        self.assertTrue(mac.is_valid("0001.0203.0405"))
        self.assertFalse(mac.is_valid("hoi"))

    def test_is_local(self):
        self.assertTrue(mac.is_local("02:00:00:00:00:00"))
        self.assertFalse(mac.is_local("00:00:00:00:00:00"))

    def test_is_multicast(self):
        self.assertTrue(mac.is_multicast("01:00:00:00:00:00"))
        self.assertFalse(mac.is_multicast("00:00:00:00:00:00"))

    def test_is_unicast(self):
        self.assertFalse(mac.is_unicast("01:00:00:00:00:00"))
        self.assertTrue(mac.is_unicast("00:00:00:00:00:00"))



#
# Copyright (c) 2024 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A vppcfg configuration module that validates sflow config"""
import logging


def validate_sflow(yaml):
    """Validate the semantics of all YAML 'sflow' config entries"""
    result = True
    msgs = []
    logger = logging.getLogger("vppcfg.config")
    logger.addHandler(logging.NullHandler())

    if not "sflow" in yaml:
        return result, msgs

    ## NOTE(pim): Nothing to validate. sflow config values are all
    ##            integers and enforced by yamale.
    return result, msgs



# Copyright (c) 2025 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""
Base class defining the interface expected by planner mixins.
"""

from abc import ABC
from typing import Dict, Any, List


class PlannerBase(ABC):  # pylint: disable=too-few-public-methods
    """Abstract base class defining the interface expected by planner mixins."""

    def __init__(self):
        # These attributes will be set by the concrete Planner class
        self.cfg: Dict[str, Any]
        self.vpp: Any  # VPPApi instance
        self.logger: Any  # Logger instance
        self.cli: Dict[str, List[str]]  # CLI commands grouped by operation type



#!/usr/bin/env python3
#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""vppcfg is a utility to configure a running VPP Dataplane using YAML
config files. See http://github.com/pimvanpelt/vppcfg/README.md for details."""
# pylint: disable=duplicate-code
import os
import sys
import logging
import yaml

# Ensure the paths are correct when we execute from the source tree
try:
    from vppcfg.config import Validator
    from vppcfg._version import __version__
except ModuleNotFoundError:
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
    from vppcfg.config import Validator
    from vppcfg._version import __version__
from vppcfg.vpp.applier import Applier
from vppcfg.vpp.planner import Planner
from vppcfg.vpp.dumper import Dumper

try:
    import argparse
except ImportError:
    print("ERROR: install argparse manually: sudo pip install argparse")
    sys.exit(-2)


def main():
    """The main vppcfg program"""
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        "-d",
        "--debug",
        dest="debug",
        action="store_true",
        help="""enable debug logging, default False""",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        help="""be quiet (only warnings/errors), default False""",
    )
    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        help="""force progress despite warnings, default False""",
    )

    subparsers = parser.add_subparsers(dest="command")
    check_p = subparsers.add_parser(
        "check", help="check given YAML config for validity (no VPP)"
    )
    check_p.add_argument(
        "-s",
        "--schema",
        dest="schema",
        type=str,
        help="""YAML schema validation file, default to use built-in""",
    )
    check_p.add_argument(
        "-c",
        "--config",
        dest="config",
        required=True,
        type=str,
        help="""YAML configuration file for vppcfg""",
    )

    dump_p = subparsers.add_parser(
        "dump", help="dump current running VPP configuration (VPP readonly)"
    )
    dump_p.add_argument(
        "-o",
        "--output",
        dest="outfile",
        required=False,
        default="-",
        type=str,
        help="""Output file for YAML config, default stdout""",
    )
    dump_p.add_argument(
        "-j",
        "--vpp-json-dir",
        dest="vpp_json_dir",
        required=False,
        type=str,
        help="""Directory where VPP API JSON files are located""",
    )
    dump_p.add_argument(
        "-a",
        "--vpp-api-socket",
        dest="vpp_api_socket",
        required=False,
        type=str,
        help="""Pathname of VPP API socket file""",
    )

    plan_p = subparsers.add_parser(
        "plan",
        help="plan changes from current VPP dataplane to target config (VPP readonly)",
    )
    plan_p.add_argument(
        "-s",
        "--schema",
        dest="schema",
        type=str,
        help="""YAML schema validation file, default to use built-in""",
    )
    plan_p.add_argument(
        "-c",
        "--config",
        dest="config",
        required=True,
        type=str,
        help="""YAML configuration file for vppcfg""",
    )
    plan_p.add_argument(
        "--novpp",
        dest="novpp",
        action="store_true",
        help="""Don't query VPP API, assume 'empty' dataplane config""",
    )
    plan_p.add_argument(
        "-o",
        "--output",
        dest="outfile",
        required=False,
        default="-",
        type=str,
        help="""Output file for VPP CLI commands, default stdout""",
    )
    plan_p.add_argument(
        "-j",
        "--vpp-json-dir",
        dest="vpp_json_dir",
        required=False,
        type=str,
        help="""Directory where VPP API JSON files are located""",
    )
    plan_p.add_argument(
        "-a",
        "--vpp-api-socket",
        dest="vpp_api_socket",
        required=False,
        type=str,
        help="""Pathname of VPP API socket file""",
    )

    apply_p = subparsers.add_parser(
        "apply", help="apply changes from current VPP dataplane to target config"
    )
    apply_p.add_argument(
        "-s",
        "--schema",
        dest="schema",
        type=str,
        help="""YAML schema validation file, default to use built-in""",
    )
    apply_p.add_argument(
        "-c",
        "--config",
        dest="config",
        required=True,
        type=str,
        help="""YAML configuration file for vppcfg""",
    )
    apply_p.add_argument(
        "-j",
        "--vpp-json-dir",
        dest="vpp_json_dir",
        required=False,
        type=str,
        help="""Directory where VPP API JSON files are located""",
    )
    apply_p.add_argument(
        "-a",
        "--vpp-api-socket",
        dest="vpp_api_socket",
        required=False,
        type=str,
        help="""Pathname of VPP API socket file""",
    )

    args = parser.parse_args()
    if not args.command:
        parser.print_help()
        print("\nPlease see vppcfg <command> -h   for per-command arguments")
        sys.exit(0)

    level = logging.INFO
    if args.debug:
        level = logging.DEBUG
    if args.quiet:
        level = logging.WARNING
    logging.basicConfig(
        format="[%(levelname)-8s] %(name)s.%(funcName)s: %(message)s", level=level
    )
    logging.info(f"vppcfg version {__version__}")

    opt_kwargs = {}
    if "vpp_json_dir" in args and args.vpp_json_dir is not None:
        opt_kwargs["vpp_json_dir"] = args.vpp_json_dir
    if "vpp_api_socket" in args and args.vpp_api_socket is not None:
        opt_kwargs["vpp_api_socket"] = args.vpp_api_socket

    if args.command == "dump":
        dumper = Dumper(**opt_kwargs)
        if not dumper.readconfig():
            logging.error("Could not retrieve config from VPP")
            sys.exit(-7)
        dumper.write(args.outfile)
        sys.exit(0)

    try:
        with open(args.config, "r", encoding="utf-8") as file:
            logging.info(f"Loading configfile {args.config}")
            cfg = yaml.load(file, Loader=yaml.FullLoader)
            logging.debug(f"Config: {cfg}")
    except OSError as err:
        logging.error(f"Couldn't read config from {args.config}: {err}")
        sys.exit(-1)

    validator = Validator(schema=args.schema)
    if not validator.valid_config(cfg):
        logging.error("Configuration is not valid, bailing")
        sys.exit(-2)
    logging.info("Configuration is valid")
    if args.command == "check":
        sys.exit(0)

    planner = Planner(cfg, **opt_kwargs)
    if args.command == "plan" and args.novpp:
        if not planner.vpp.mockconfig(cfg):
            sys.exit(-7)
    else:
        if not planner.vpp.readconfig():
            sys.exit(-3)

        if not planner.phys_exist_in_vpp():
            logging.error("Not all PHYs in the config exist in VPP")
            sys.exit(-4)

        if not planner.phys_exist_in_config():
            logging.error("Not all PHYs in VPP exist in the config")
            sys.exit(-5)

        if not planner.lcps_exist_with_lcp_enabled():
            logging.error(
                "Linux Control Plane is needed, but linux-cp API is not available"
            )
            sys.exit(-6)

    failed = False
    if not planner.prune():
        if not args.force:
            logging.error("Planning prune failure")
            sys.exit(-10)
        failed = True
        logging.warning("Planning prune failure, continuing due to --force")

    if not planner.create():
        if not args.force:
            logging.error("Planning create failure")
            sys.exit(-20)
        failed = True
        logging.warning("Planning create failure, continuing due to --force")

    if not planner.sync():
        if not args.force:
            logging.error("Planning sync failure")
            sys.exit(-30)
        failed = True
        logging.warning("Planning sync failure, continuing due to --force")

    if args.command == "plan":
        planner.write(args.outfile, emit_ok=not failed)

    if failed:
        logging.error("Planning failed")
        sys.exit(-40)

    logging.info("Planning succeeded")
    if args.command == "plan":
        sys.exit(0)

    if args.command == "apply":
        applier = Applier(cfg, planner.get_commands())
        if not applier.apply():
            logging.error("Applying configuration failed")
            sys.exit(-50)
        logging.info("Apply succeeded")

    sys.exit(0)


if __name__ == "__main__":
    main()
